import os
from pywapor.collect.protocol import cds
from pywapor.general.processing_functions import open_ds, remove_ds
from pywapor.general.logger import log, adjust_logger
from pywapor.enhancers.pressure import pa_to_kpa
from pywapor.enhancers.wind import adjust_wind_height, windspeed
from pywapor.enhancers.temperature import kelvin_to_celsius
from functools import partial
import numpy as  np

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {
        "sis-agrometeorological-indicators": {
                    "2m_temperature": [{"statistic": "24_hour_mean", "format": "zip", "version": "1_1"}, "t_air"],
                    "2m_dewpoint_temperature": [{"statistic": "24_hour_mean", "format": "zip", "version": "1_1"}, "t_dew"],
                    "2m_relative_humidity": [{"time": [f"{x:02d}_00" for x in [6,9,12,15,18]], "format": "zip", "version": "1_1"}, "rh"],
                    "10m_wind_speed": [{"statistic": "24_hour_mean", "format": "zip", "version": "1_1"}, "u"],
                    "vapour_pressure": [{"statistic": "24_hour_mean", "format": "zip", "version": "1_1"}, "vp"],
                    "solar_radiation_flux": [{"statistic": "24_hour_mean", "format": "zip", "version": "1_1"}, "ra"],
                        },

        "reanalysis-era5-single-levels": {
            'total_column_water_vapour':[{"time": [f"{x:02d}:00" for x in range(24)], "format": "netcdf", "product_type": "reanalysis"}, "wv"],
            '10m_u_component_of_wind':  [{"time": [f"{x:02d}:00" for x in range(24)], "format": "netcdf", "product_type": "reanalysis"}, "u10m"],
            '10m_v_component_of_wind':  [{"time": [f"{x:02d}:00" for x in range(24)], "format": "netcdf", "product_type": "reanalysis"}, "v10m"],
            '2m_dewpoint_temperature':  [{"time": [f"{x:02d}:00" for x in range(24)], "format": "netcdf", "product_type": "reanalysis"}, "t_dew"],
            'mean_sea_level_pressure':  [{"time": [f"{x:02d}:00" for x in range(24)], "format": "netcdf", "product_type": "reanalysis"}, "p_air_0"],
            'surface_pressure':         [{"time": [f"{x:02d}:00" for x in range(24)], "format": "netcdf", "product_type": "reanalysis"}, "p_air"],
            '2m_temperature':           [{"time": [f"{x:02d}:00" for x in range(24)], "format": "netcdf", "product_type": "reanalysis"}, "t_air"],
        }
    }

    req_dl_vars = {
        "sis-agrometeorological-indicators": {
            "t_air": ["2m_temperature"],
            "t_air_max": ["2m_temperature"],
            "t_air_min": ["2m_temperature"],
            "t_dew": ["2m_dewpoint_temperature"],
            "rh": ["2m_relative_humidity"],
            "u": ["10m_wind_speed"],
            "vp": ["vapour_pressure"],
            "ra": ["solar_radiation_flux"],
        },
        "reanalysis-era5-single-levels": {
            "wv": ['total_column_water_vapour'],
            "u10m": ['10m_u_component_of_wind'],
            "v10m": ['10m_v_component_of_wind'],
            "u2m": ['10m_u_component_of_wind'],
            "v2m": ['10m_v_component_of_wind'],
            "u": ['10m_u_component_of_wind', '10m_v_component_of_wind'],
            "t_dew": ['2m_dewpoint_temperature'],
            "p_air_0": ['mean_sea_level_pressure'],
            "p_air": ['surface_pressure'],
            "t_air": ['2m_temperature'],
            "t_air_min": ['2m_temperature'],
            "t_air_max": ['2m_temperature'],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}

    return out

def jouleperday_to_watt(ds, var):
    ds[var] = ds[var] / 86400
    return ds

def default_post_processors(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "sis-agrometeorological-indicators": {
            "t_air": [kelvin_to_celsius],
            "t_air_max": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_max")],
            "t_air_min": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_min")],
            "t_dew": [kelvin_to_celsius],
            "rh": [],
            "u": [adjust_wind_height],
            "vp": [], #is already mbar
            "ra": [jouleperday_to_watt],
        },
        "reanalysis-era5-single-levels": {
            "wv": [], # is already kg/m2
            "u10m": [],
            "v10m": [],
            "u2m": [adjust_wind_height],
            "v2m": [adjust_wind_height],
            "u": [windspeed, adjust_wind_height],
            "t_dew": [kelvin_to_celsius],
            "p_air_0": [pa_to_kpa],
            "p_air": [pa_to_kpa],
            "t_air": [kelvin_to_celsius],
            "t_air_max": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_max")],
            "t_air_min": [partial(kelvin_to_celsius, in_var = "t_air", out_var = "t_air_min")],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def download(folder, latlim, lonlim, timelim, product_name, req_vars, 
                variables = None, post_processors = None):
    """Download ERA5 data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    product_folder = os.path.join(folder, "ERA5")

    if not os.path.exists(product_folder):
        os.makedirs(product_folder)

    fn_final = os.path.join(product_folder, f"{product_name}.nc")
    if os.path.isfile(fn_final):
        ds = open_ds(fn_final)
        if np.all([x in ds.data_vars for x in req_vars]):
            return ds[req_vars]
        else:
            remove_ds(ds)

    spatial_buffer = True
    if spatial_buffer:
        latlim = [latlim[0] - 0.1, latlim[1] + 0.1]
        lonlim = [lonlim[0] - 0.1, lonlim[1] + 0.1]

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items()}

    ds = cds.download(product_folder, product_name, latlim, lonlim, timelim, variables, post_processors)

    return ds

if __name__ == "__main__":

    folder = r"/Users/hmcoerver/Local/era_test"
    latlim = [28.9, 29.7]
    lonlim = [30.2, 31.2]
    timelim = ["2021-06-26", "2021-07-05"]

    # product_name = "sis-agrometeorological-indicators"
    product_name = "reanalysis-era5-single-levels"
    req_vars = ["t_air", "t_dew"]#, "vp", "ra"]
    # req_vars = ["u"]

    variables = None
    post_processors = None

    adjust_logger(True, folder, "INFO")

    ds = download(folder, latlim, lonlim, timelim, product_name = product_name, 
                req_vars = req_vars, variables = variables, post_processors = post_processors)



"""
https://github.com/nsidc/earthaccess/discussions/489
https://search.earthdata.nasa.gov/search/granules/collection-details?p=C2105091501-LAADS&pg[0][v]=f&pg[0][gsk]=-start_date&q=VNP02IMG&tl=1689845353!3!!
"""

import requests
import datetime
import base64
import urllib
import copy
import os
import multiprocessing
import xarray as xr
import numpy as np
import tqdm
from osgeo import gdal
from functools import partial
from pywapor.collect import accounts
from pywapor.collect.protocol.crawler import download_urls, download_url
from joblib import Parallel, delayed, Memory
from pywapor.enhancers.apply_enhancers import apply_enhancers
from pywapor.general.processing_functions import save_ds, remove_ds, adjust_timelim_dtype, is_corrupt_or_empty, open_ds
from pywapor.general.logger import log, adjust_logger
from pywapor.enhancers.other import drop_empty_times
from pywapor.general.curvilinear import curvi_to_recto, create_grid
from pywapor.collect.protocol.opendap import make_opendap_url
gdal.UseExceptions()

def get_token():
    s3_endpoint = "https://data.laadsdaac.earthdatacloud.nasa.gov/s3credentials"
    
    login_resp = requests.get(s3_endpoint, allow_redirects=False)
    login_resp.raise_for_status()

    un, pw = accounts.get("NASA")
    auth = f"{un}:{pw}"
    encoded_auth  = base64.b64encode(auth.encode('ascii'))

    auth_redirect = requests.post(
        login_resp.headers['location'],
        data = {"credentials": encoded_auth},
        headers= {"Origin": s3_endpoint},
        allow_redirects=False
    )
    auth_redirect.raise_for_status()
    if "currently unavailable" in auth_redirect.text:
        log.warning(f"The Earthdata Service is currently unavailable. Check {login_resp.headers['location']}.")
        raise ConnectionError(f"Earthdata Service is currently unavailable. Check {login_resp.headers['location']}.")

    final = requests.get(auth_redirect.headers['location'], allow_redirects=False)
    results = requests.get(s3_endpoint, cookies={'accessToken': final.cookies['accessToken']})
    results.raise_for_status()

    return results.json()

def adj(number, make):
    """
    Check if a number is odd or even and adjust it
    if necessary.
    """
    is_even = number % 2 == 0
    if is_even:
        if make == "odd_up":
            number += 1
        if make == "odd_down":
            number -= 1
    else:
        if make == "even_down":
            number -= 1
        if make == "even_up":
            number += 1
    return int(number)

def download_arrays(path, subdss, folder, path_appendix = ".nc", creation_options = ["ARRAY:COMPRESS=DEFLATE"], dtype_is_8bit = False):

    fn = os.path.splitext(os.path.split(path)[-1])[0]
    path_local = os.path.join(folder, fn + path_appendix)

    if os.path.isfile(path_local):
        remove_ds(path_local)

    # check if file is local or vsicurl or vsis3
    local = not urllib.parse.urlparse(path).scheme in ('http', 'https',)

    if (not local) and ("vsis3" in path):
        creds = get_token()

        gdal_config_options = {
            "AWS_ACCESS_KEY_ID": creds["accessKeyId"],
            "AWS_SESSION_TOKEN": creds["sessionToken"],
            "AWS_SECRET_ACCESS_KEY": creds["secretAccessKey"],
            "GDAL_DISABLE_READDIR_ON_OPEN": "EMPTY_DIR",
        }
    else:
        gdal_config_options = {}

    try:
        for k, v in gdal_config_options.items():
            gdal.SetConfigOption(k, v)

        md_options = gdal.MultiDimTranslateOptions(
                arraySpecs = subdss,
                creationOptions = creation_options
        )

        ds = gdal.MultiDimTranslate(path_local, path, options = md_options)
        ds = ds.FlushCache()
    except Exception as e:
        raise e
    finally:
        for k, v in gdal_config_options.items():
            gdal.SetConfigOption(k, None)

    return path_local

def search_stac(params, cachedir = None, extra_filters = {"day_night_flag": "DAY"}):
    
    memory = Memory(cachedir, verbose=0)

    @memory.cache()
    def _post_search(search, params_):
        query = requests.post(search, json = params_)
        query.raise_for_status()
        out = query.json()
        return out

    @memory.cache()
    def _check(ftr, extra_filters):
        links = [x for x in ftr.get("links", []) if x["rel"] == "via"]
        link = links[0].get("href", None) if len(links) > 0 else None
        metadata_resp = requests.get(link)
        metadata_resp.raise_for_status()
        metadata = metadata_resp.json()
        check = [metadata[k] == v for k,v in extra_filters.items()]
        return all(check)

    search = 'https://cmr.earthdata.nasa.gov/stac/LAADS/search'
    all_scenes = list()
    links = [{"body": params, "rel": "next"}]
    while "next" in [x.get("rel", None) for x in links]:
        params_ = [x.get("body") for x in links if x.get("rel") == "next"][0]
        out = _post_search(search, params_)
        all_scenes += out["features"]
        links = out["links"]
    log.info(f"--> Found {len(all_scenes)} `{params_['collections'][0]}` scenes.")

    if len(extra_filters) > 0:
        log.add().info(f"--> Filtering `{params_['collections'][0]}` scenes with `{extra_filters}`.")
        final = [x for x in tqdm.tqdm(all_scenes, position = 0, bar_format='{l_bar}{bar}|', delay = 5, leave = False) if _check(x, extra_filters)]
        log.info(f"--> Found {len(final)} relevant `{params_['collections'][0]}` scenes.").sub()
    else:
        final = all_scenes

    return final

def create_stac_summary(bb, timelim, cachedir=None):
    
    sd = datetime.datetime.strftime(timelim[0], "%Y-%m-%dT00:00:00Z")
    ed = datetime.datetime.strftime(timelim[1], "%Y-%m-%dT23:59:59Z")
    search_dates = f"{sd}/{ed}"

    params = dict()
    params['limit'] = 250
    params['bbox'] = bb
    params['datetime'] = search_dates

    params['collections'] = ['VNP03IMG.v2']
    results03 = search_stac(params, cachedir=cachedir, extra_filters={})
    params['collections'] = ['CLDMSK_L2_VIIRS_SNPP.v1']
    resultsqa = search_stac(params, cachedir=cachedir, extra_filters={})
    params['collections'] = ['VNP02IMG.v2']
    results02 = search_stac(params, cachedir=cachedir)

    get_fns = lambda results: [os.path.split(x["assets"]["data"]["href"])[-1] for x in results]
    get_dts = lambda files: {datetime.datetime.strptime(".".join(x.split(".")[1:3]), "A%Y%j.%H%M"): x for x in files}
    files02 = get_fns(results02)
    files03 = get_fns(results03)
    filesqa = get_fns(resultsqa)

    dates02 = get_dts(files02)
    dates03 = get_dts(files03)
    datesqa = get_dts(filesqa)

    unique_dates = set(dates02.keys()).union(set(dates03.keys())).union(set(datesqa.keys()))

    summary = {k: (dates02.get(k, None), dates03.get(k, None), datesqa.get(k, None)) for k in unique_dates}
    filtered_summary = {k: v for k, v in summary.items() if not None in  v}

    return filtered_summary

def default_vars(product_name, req_vars):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """

    variables = {
        "VNP02IMG": {"bt": {(), "bt"}},
    }

    req_dl_vars = {
        "VNP02IMG": {
            "bt": ["bt"],
        },
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def default_post_processors(product_name, req_vars = None):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list
        List of variables to be collected.

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """
    post_processors = {
        "VNP02IMG": {
            "bt": [partial(drop_empty_times, drop_vars = ["bt"])]
            },
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def preproc(ds):
    ds = ds.rename({"lat": "y", "lon": "x", "Band1": "bt"})
    ds = ds.drop("crs")
    ds.attrs = {}
    for var in ds.data_vars:
        ds[var].attrs = {}
    ds = ds.rio.write_grid_mapping("spatial_ref")
    ds = ds.rio.write_crs("epsg:4326")
    ds = ds.rio.write_transform(ds.rio.transform(recalc=True))
    fn = os.path.split(ds.encoding["source"])[-1]
    date = np.datetime64(fn[3:13] + " " + fn[14:22].replace("_", ":"), "ns")
    return ds.expand_dims("time").assign_coords({"time": [date]})

def combine_unprojected_data(nc02_file, ncqa_file, lut_file, unproj_fn):

    if isinstance(lut_file, type(None)):
        ds_ = xr.open_dataset(nc02_file, mask_and_scale=False, group = "observation_data")
        ds1 = ds_[["I05", "I05_quality_flags"]]
        ds2 = ds_[["I05_brightness_temperature_lut"]]
        ds3 = xr.open_dataset(ncqa_file, chunks = "auto", group = "geophysical_data")
    else:
        ds1 = xr.open_dataset(nc02_file, mask_and_scale=False)
        ds2 = xr.open_dataset(lut_file, mask_and_scale=False)
        ds3 = xr.open_dataset(ncqa_file, chunks = "auto")

    # Create cloud mask. (0=cloudy, 1=probably cloudy, 2=probably clear, 3=confident clear, -1=no result)
    cmask = ds3["Integer_Cloud_Mask"].interp({
                    "number_of_lines": np.linspace(0-0.25, (ds3["number_of_lines"].size-1)+0.25, ds3["number_of_lines"].size*2),
                    "number_of_pixels": np.linspace(0-0.25, (ds3["number_of_pixels"].size-1)+0.25, ds3["number_of_pixels"].size*2)
                    }, kwargs = {"fill_value": "extrapolate"}).drop_vars(["number_of_lines", "number_of_pixels"])

    # Convert DN to Brightness Temperature using the provided lookup table.
    bt_da = ds2["I05_brightness_temperature_lut"].isel(number_of_LUT_values = ds1["I05"])

    # Chunk and mask invalid pixels.
    bt = bt_da.chunk("auto").where((bt_da >= bt_da.attrs["valid_min"]) & 
                                        (bt_da <= bt_da.attrs["valid_max"]) & 
                                        (bt_da != bt_da.attrs["_FillValue"]) &
                                        (ds1["I05_quality_flags"] == 0) &
                                        (cmask == 3), bt_da.attrs["_FillValue"]
    ).rename("bt").to_dataset()

    _ = save_ds(bt, unproj_fn, encoding = "initiate", label = "Combining data.").close()

def download(folder, latlim, lonlim, timelim, product_name, req_vars,
                variables = None, post_processors = None):
    """Download VIIRSL1 data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list
        Which variables to download for the selected product.
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "VIIRSL1")
    cachedir = os.path.join(folder, "cache")
    if not os.path.exists(folder):
        os.makedirs(folder)

    fn = os.path.join(folder, f"{product_name}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]
        
    timelim = adjust_timelim_dtype(timelim)

    spatial_buffer = True
    if spatial_buffer:
        dx = dy = 0.0033
        latlim = [latlim[0] - dy, latlim[1] + dy]
        lonlim = [lonlim[0] - dx, lonlim[1] + dx]

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    bb, nx, ny = create_grid(latlim, lonlim, dx_dy = (0.0033, 0.0033))
    filtered_summary = create_stac_summary(bb, timelim, cachedir = cachedir)

    log.info(f"--> Collecting {len(filtered_summary)} VIIRS scenes.")

    buckets = ("VNP02IMG", "VNP03IMG","CLDMSK_L2_VIIRS_SNPP")
    all_urls = {k: tuple(f"/vsis3/prod-lads/{buckets[i]}/{v[i]}" for i in range(3)) for k, v in filtered_summary.items()}
    all_proj_files = list()

    for i, (date, (nc02, nc03, nc_cloud)) in enumerate(copy.deepcopy(all_urls).items()):
        date_str = str(date).replace(" ", "_").replace(":","_")
        proj_fn = os.path.join(folder, f"bt_{date_str}_projected.nc")
        if os.path.isfile(proj_fn):
            all_proj_files.append(proj_fn)
            _ = all_urls.pop(date)

    log.info(f"--> {len(all_proj_files)} scenes already downloaded, {len(all_urls)} remaining.").add()

    dl_method = "opendap"
    if dl_method == "get_request":
        token = accounts.get("VIIRSL1")[1]

    # i = 0
    # date = list(all_urls.keys())[i]
    # nc02, nc03, nc_cloud = all_urls[date]

    for i, (date, (nc02, nc03, nc_cloud)) in enumerate(all_urls.items()):

        date_str = str(date).replace(" ", "_").replace(":","_")
        unproj_fn = os.path.join(folder, f"bt_{date_str}_unprojected.nc")
        proj_fn = os.path.join(folder, f"bt_{date_str}_projected.nc")

        log.info(f"--> ({i+1}/{len(all_urls)}) Processing '{os.path.split(nc02)[-1]}'.").add()

        cleanup = []

        if dl_method == "s3_bucket": # NOTE only works in-region, see https://github.com/nsidc/earthaccess/discussions/489
            log.warning("--> The `s3_bucket` method only works if you are running this script from the correct region.")
            parallel = True
            if parallel:
                inputs = [
                    {"path": nc03, "subdss": ["/geolocation_data/latitude"], "folder": folder, "path_appendix": "_lat.nc"},
                    {"path": nc03, "subdss": ["/geolocation_data/longitude"], "folder": folder, "path_appendix": "_lon.nc"},
                    {"path": nc02, "subdss": ["/observation_data/I05_quality_flags", "/observation_data/I05"], "folder": folder},
                    {"path": nc02, "subdss": ["/observation_data/I05_brightness_temperature_lut"], "folder": folder, "path_appendix": "_lut.nc"},
                    {"path": nc_cloud, "subdss": ["/geophysical_data/Integer_Cloud_Mask"], "folder": folder, "creation_options":["COMPRESS=DEFLATE"], "dtype_is_8bit": True},
                ]
                n_jobs = min(5, multiprocessing.cpu_count())
                lats_file, lons_file, nc02_file, lut_file, ncqa_file = Parallel(n_jobs=n_jobs)(delayed(download_arrays)(**i) for i in inputs)
            else:
                lats_file = download_arrays(nc03, ["/geolocation_data/latitude"], folder, path_appendix="_lat.nc")
                lons_file = download_arrays(nc03, ["/geolocation_data/longitude"], folder, path_appendix="_lon.nc")
                nc02_file = download_arrays(nc02, ["/observation_data/I05_quality_flags", "/observation_data/I05"], folder)
                lut_file = download_arrays(nc02, ["/observation_data/I05_brightness_temperature_lut"], folder, path_appendix = "_lut.nc")
                ncqa_file = download_arrays(nc_cloud, ["/geophysical_data/Integer_Cloud_Mask"], folder, creation_options=["COMPRESS=DEFLATE"], dtype_is_8bit=True)

        elif dl_method == "get_request":
            base_url = "https://ladsweb.modaps.eosdis.nasa.gov/archive/allData/"
            nc02_parts = os.path.normpath(nc02).split(os.sep)
            nc03_parts = os.path.normpath(nc03).split(os.sep)
            nc_cloud_parts = os.path.normpath(nc_cloud).split(os.sep)
            year_doy = nc02_parts[-1].split(".")[1][1:]
            nc02_url = os.path.join(base_url, "5200", nc02_parts[-2], year_doy[:4], year_doy[4:], nc02_parts[-1])
            nc03_url = os.path.join(base_url, "5200", nc03_parts[-2], year_doy[:4], year_doy[4:], nc03_parts[-1])
            ncqa_url = os.path.join(base_url, "5110", nc_cloud_parts[-2], year_doy[:4], year_doy[4:], nc_cloud_parts[-1])
            urls = [nc02_url, nc03_url, ncqa_url]
            groups = ["observation_data", "geolocation_data", "geophysical_data"]
            for group, url in zip(groups, urls.copy()):
                fp_ = os.path.join(folder, os.path.split(url)[-1])
                if os.path.isfile(fp_):
                    corrupt = is_corrupt_or_empty(fp_, group = group)
                    if corrupt:
                        remove_ds(fp_)
                    else:
                        log.info(f"--> Skipping downloading of `{os.path.split(url)[-1]}`.")
                        urls.remove(url)
            headers = {'Authorization': 'Bearer ' + token}
            n_jobs = min(len(urls), multiprocessing.cpu_count())
            _ = download_urls(urls, folder, headers = headers, parallel = n_jobs)
            nc02_file_ = os.path.join(folder, os.path.split(nc02_url)[-1])
            nc03_file = os.path.join(folder, os.path.split(nc03_url)[-1])
            ncqa_file_ = os.path.join(folder, os.path.split(ncqa_url)[-1])
            # NOTE The data has already been downloaded, just running this to align the files
            # with the `method = s3_bucket` branch.
            lats_file = download_arrays(nc03_file, ["/geolocation_data/latitude"], folder, path_appendix="_lat.nc")
            lons_file = download_arrays(nc03_file, ["/geolocation_data/longitude"], folder, path_appendix="_lon.nc")
            lut_file = download_arrays(nc02_file_, ["/observation_data/I05_brightness_temperature_lut"], folder, path_appendix = "_lut.nc")
            nc02_file = download_arrays(nc02_file_, ["/observation_data/I05_quality_flags", "/observation_data/I05"], folder, path_appendix = "_data.nc")
            ncqa_file = download_arrays(ncqa_file_, ["/geophysical_data/Integer_Cloud_Mask"], folder, creation_options=["COMPRESS=DEFLATE"], path_appendix = "_cloud.nc", dtype_is_8bit=True)
            cleanup += [nc03_file, nc02_file_, ncqa_file_]

        elif dl_method == "opendap":

            base_url = "https://ladsweb.modaps.eosdis.nasa.gov/opendap/RemoteResources/laads/allData"
            nc02_parts = os.path.normpath(nc02).split(os.sep)
            nc03_parts = os.path.normpath(nc03).split(os.sep)
            nc_cloud_parts = os.path.normpath(nc_cloud).split(os.sep)
            year_doy = nc02_parts[-1].split(".")[1][1:]
            base_url02 = os.path.join(base_url, "5200", nc02_parts[-2], year_doy[:4], year_doy[4:], nc02_parts[-1])
            base_url03 = os.path.join(base_url, "5200", nc03_parts[-2], year_doy[:4], year_doy[4:], nc03_parts[-1])
            base_cloud = os.path.join(base_url, "5110", nc_cloud_parts[-2], year_doy[:4], year_doy[4:], nc_cloud_parts[-1])

            # Download geolocation data.
            order = {
                "/geolocation_data/longitude": {},
                "/geolocation_data/latitude": {},
            }
            url = make_opendap_url(base_url03, order)
            log.info(f"--> Downloading `{nc03_parts[-1]}`.")
            nc03_file = download_url(url, os.path.join(folder, nc03_parts[-1]))

            # Determine AOI inside scene.
            log.info(f"--> Determining indices of AOI inside scene.").add()
            geoloc_ = xr.open_dataset(nc03_file, group = "geolocation_data")
            coords = {
                **{f"{k}_750": (k, np.repeat(np.arange(0, v/2), 2)) for k, v in geoloc_.sizes.items()},
                **{f"{k}_375": (k, np.arange(0, v)) for k, v in geoloc_.sizes.items()},
            }
            geoloc = geoloc_.assign_coords(coords)
            geoloc = geoloc.where(
                        (geoloc.longitude < lonlim[1]) &
                        (geoloc.longitude > lonlim[0]) &
                        (geoloc.latitude > latlim[0]) &
                        (geoloc.latitude < latlim[1]), drop = True
                        )
            domain_375 = {k.replace("_375", ""): [adj(geoloc[k].min(), "even_down"), adj(geoloc[k].max(), "odd_up")] for k in geoloc.coords if "_375" in k}
            domain_750 = {k.replace("_750", ""): [int(geoloc[k].min()), int(geoloc[k].max())] for k in geoloc.coords if "_750" in k}
            log.info(f"--> Found {np.prod([v[1] - v[0] + 1 for v in domain_375.values()]):,} pixels inside AOI.").sub()

            geoloc_ds = geoloc_.isel({k: slice(v[0], v[1]+1) for k, v in domain_375.items()})
            lats_file_ds = save_ds(geoloc_ds[["latitude"]], os.path.join(folder, nc03_parts[-1].replace(".nc", "_lat.nc")), label = "Saving latitudes.")
            lats_file = lats_file_ds.encoding["source"]
            lats_file_ds.close()
            
            lons_file_ds = save_ds(geoloc_ds[["longitude"]], os.path.join(folder, nc03_parts[-1].replace(".nc", "_lon.nc")), label = "Saving longitudes.")
            lons_file = lons_file_ds.encoding["source"]
            lons_file_ds.close()

            # # Check domains
            # for dim in list(domain_750.keys()):
            #     size_750 = domain_750[dim][1] - domain_750[dim][0] + 1
            #     size_375 = domain_375[dim][1] - domain_375[dim][0] + 1
            #     print(dim, size_750, size_375)
            #     print(size_750 == size_375/2)

            # Download observation data.
            log.info(f"--> Downloading `{nc02_parts[-1]}` for AOI.")
            order = {
                "/observation_data/I05_quality_flags": domain_375,
                "/observation_data/I05": domain_375,
                "/observation_data/I05_brightness_temperature_lut": {}
            }
            url = make_opendap_url(base_url02, order)
            nc02_file = download_url(url, os.path.join(folder, nc02_parts[-1].replace(".nc", "_data_lut.nc")))

            # Download cloud mask.
            log.info(f"--> Downloading `{nc_cloud_parts[-1]}` for AOI.")
            order = {
                "/geophysical_data/Integer_Cloud_Mask": domain_750,
            }
            url = make_opendap_url(base_cloud, order)
            ncqa_file = download_url(url, os.path.join(folder, nc_cloud_parts[-1].replace(".nc", "_cloud.nc")))

            lut_file = None
            cleanup += [nc03_file]
        else:
            raise ValueError

        if not os.path.isfile(unproj_fn):
            combine_unprojected_data(nc02_file, ncqa_file, lut_file, unproj_fn)

        warp_kwargs = {"outputBounds": bb, "width": nx, "height": ny}
        _ = curvi_to_recto(lats_file, lons_file, {"bt": unproj_fn}, proj_fn, warp_kwargs = warp_kwargs)
        all_proj_files.append(proj_fn)

        cleanup += [nc02_file, ncqa_file, lut_file, unproj_fn, lats_file, lons_file]
        for x in cleanup:
            if not isinstance(x, type(None)):
                remove_ds(x)

        log.sub()

    log.sub()

    ds = xr.open_mfdataset(all_proj_files, preprocess = preproc)

    # Apply product specific functions.
    ds = apply_enhancers(post_processors, ds)

    ds = save_ds(ds, fn, encoding = "initiate", label = "Merging files.")

    return ds[req_vars_orig]

if __name__ == "__main__":

    folder = "/Users/hmcoerver/Local/test_dl_VIIRSL1_0"
    latlim = [29.4, 29.7]
    lonlim = [30.7, 31.0]
    timelim = [datetime.date(2023, 3, 1), datetime.date(2023, 3, 2)]
    product_name = "VNP02IMG"

    variables = None
    post_processors = None

    req_vars = ["bt"]

    adjust_logger(True, folder, "INFO")









    # timelim = [np.datetime64 ('2023-07-03T00:00:00.000000000'), 
    #            np.datetime64('2023-07-10T00:00:00.000000000')]
    # latlim = [29.4, 29.6]
    # lonlim = [30.7, 30.9]

    # timelim = [datetime.datetime(2021, 3, 4), datetime.datetime(2021, 5, 5)]
    # lonlim = [22.78125, 32.48438]
    # latlim = [23.72777, 32.1612]
    # # folder = workdir = r"/Users/hmcoerver/Local/viirs"
    # product_name = "VNP02IMG"
    # req_vars = ["bt"]
    # variables = None
    # post_processors = None
    # folder = r"/Users/hmcoerver/Local/viirs_test"

    # adjust_logger(True, folder, "INFO")

    # i = 2

    # folder = os.path.join(folder, "VIIRSL1")
    # date = datetime.datetime(2022, 4, 30, 7, 54)
    # nc02 = '/vsis3/prod-lads/VNP02IMG/VNP02IMG.A2022120.0754.002.2022270171852.nc'
    # nc03 = '/vsis3/prod-lads/VNP03IMG/VNP03IMG.A2022120.0754.002.2022120162348.nc'
    # nc_cloud = '/vsis3/prod-lads/CLDMSK_L2_VIIRS_SNPP/CLDMSK_L2_VIIRS_SNPP.A2022120.0754.001.2022120192728.nc'

    # path, subdss, folder = (nc03, ["/geolocation_data/latitude"], folder)
    # path_appendix="_lat.nc"

    # creation_options = ["ARRAY:COMPRESS=DEFLATE"]
    # dtype_is_8bit = False

    # bb, nx, ny = create_grid(latlim, lonlim, dx_dy = (0.0033, 0.0033))
    # timelim = adjust_timelim_dtype(timelim)

    # params = {'limit': 250,
    #             'bbox': [22.78125, 23.72777, 32.48655, 32.16257],
    #             'datetime': '2021-03-04T00:00:00Z/2021-05-05T23:59:59Z',
    #             'collections': ['VNP02IMG.v2']}

    # endpoint = 'https://cmr.earthdata.nasa.gov/stac/LAADS'
    # extra_filters = {"day_night_flag": "DAY"}

    # # cachedir = 'your_cache_location_directory'
    # cachedir = os.path.join(folder, "VIIRS_API_cache")

    # # memory = Memory(cachedir, verbose=0)


import os
import rasterio
from rasterio import shutil as rio_shutil
from rasterio.vrt import WarpedVRT
from rasterio.crs import CRS
import xarray as xr
import numpy as np
import os
from pywapor.general.logger import log
from pywapor.general.performance import performance_check
from pywapor.general.processing_functions import process_ds
from pywapor.general.processing_functions import open_ds, save_ds, remove_ds, log_example_ds
import rasterio
import warnings

def get_pixel_sizes(dss):
    """Determines to pixel size for each dataset in `dss`.

    Parameters
    ----------
    dss : list
        List of xr.Datasets.

    Returns
    -------
    dict
        Dictionary where keys are the pixel area and values the datasets.
    """
    # Check CRSs of datasets.
    crss = [v.rio.crs.to_epsg() for v in dss]
    # Count occurence of the different CRSs.
    uniqs, counts = np.unique(crss, return_counts=True)
    # Pick the dominant CRS.
    crs = uniqs[np.argmax(counts)]
    # Reproject to common CRS.
    dss2 = [(ds, ds.rio.reproject(CRS.from_epsg(crs))) for ds in dss]
    return {np.abs(np.prod(ds2.rio.resolution())): ds for ds, ds2 in dss2}

def has_changed(ds):
    """Checks if a dataset has been changed since it was opened from disk.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to check.

    Returns
    -------
    bool
        `True` if the dataset has changed since it was opened. 
    """
    if "source" in ds.encoding.keys():
        changed = not ds.equals(open_ds(ds.encoding["source"]))
    else:
        changed = True
    return changed

def has_geotransform(ds):
    """Checks if a dataset has a geotransform that can be read by `rasterio`.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to check.

    Returns
    -------
    bool
        `True` if the dataset has a correctly set geotransform.
    """
    varis = ds.data_vars
    if not "source" in ds.encoding.keys():
        return False
    for var in varis:
        with warnings.catch_warnings(record=True) as w:
            _ = rasterio.open(f"netcdf:{ds.encoding['source']}:{var}")
            if len(w) > 0:
                for warning in w:
                    no_geot = "Dataset has no geotransform, gcps, or rpcs." in str(warning.message)
                    if no_geot:
                        return False
    return True

def align_pixels(dss, folder, spatial_interp = "nearest", example_ds = None, stack_dim = "time", fn_append = ""):
    """Aligns pixels (spatially) among the given datasets.

    Parameters
    ----------
    dss : list
        List of datasets that need to be spatially aligned.
    folder : str
        Path to folder in which to store new files.
    spatial_interp : str, optional
        Spatial interpolation to use, by default "nearest".
    example_ds : xr.Dataset, optional
        Dataset to use as example, if None a example dataset is chosen by selectiona dataset from `dss`
        with the smallest pixels, by default None.
    stack_dim : str, optional
        Dimension name that is not spatial, by default "time".
    fn_append : str, optional
        String that will be appended to any files created, by default "".

    Returns
    -------
    tuple
        First value contains a list with all the aligned datasets, second value contains a list with any new files
        that have been created.
    """

    temp_files = list()

    if isinstance(spatial_interp, str):
        spatial_interp = [spatial_interp] * len(dss)
    assert len(dss) == len(spatial_interp)

    if len(dss) == 1 and isinstance(example_ds, type(None)):
        dss1 = [list(dss)[0]]
    else:
        if isinstance(example_ds, type(None)):
            example_ds = min(get_pixel_sizes(dss).items(), key=lambda x: x[0])[1]
            log_example_ds(example_ds)
        dss1 = list()
        for i, (spat_interp, ds_part) in enumerate(zip(spatial_interp, dss)):
            if needs_reprojecting(ds_part, example_ds):
                var_str = "_".join(ds_part.data_vars)
                dst_path = os.path.join(folder, f"{var_str}_x{i}{fn_append}.nc")
                ds_part = reproject(ds_part, example_ds, dst_path, spatial_interp = spat_interp, stack_dim = stack_dim)
                temp_files.append(ds_part)
            else:
                ds_part = ds_part.assign_coords({"y": example_ds.y, 
                                                "x": example_ds.x, 
                                                "spatial_ref": example_ds.spatial_ref})
            dss1.append(ds_part)

    return dss1, temp_files

def needs_reprojecting(ds, example_ds):
    """Check if a dataset is already aligned with `example_ds` or not.

    Parameters
    ----------
    ds : xr.Dataset
        Dataset to check.
    example_ds : xr.Dataset
        Dataset to check against.

    Returns
    -------
    bool
        whether or not the `ds` needs to be reprojected.
    """
    check_crs = ds.rio.crs == example_ds.rio.crs
    check_ysize = ds.y.size == example_ds.y.size
    check_xsize = ds.x.size == example_ds.x.size
    check_y = bool((ds.y == example_ds.y).all().values)
    check_x = bool((ds.x == example_ds.x).all().values)
    return not np.all([check_crs, check_ysize, check_xsize, check_x, check_y])

def choose_reprojecter(src_ds, max_bytes = 2e9, min_times = 10, stack_dim = "time"):
    """Choose which reprojecting function to use (`reproject_bulk` or `reproject_chunk`).

    Parameters
    ----------
    src_ds : xr.Dataset
        Dataset to be reprojected.
    max_bytes : int, optional
        Threshold value to switch to chunked reprojecting, by default 2e9.
    min_times : int, optional
        Threshold value to switch to bulk reprojecting, by default 10.
    stack_dim : str, optional
        Dimension name that is not spatial, by default "time".

    Returns
    -------
    function
        Function to use for reprojection.
    """

    if stack_dim in src_ds.dims:
        tsize = src_ds.sizes[stack_dim]
    else:
        tsize = 1

    if src_ds.nbytes < max_bytes and tsize > min_times:
        reproject = reproject_bulk
    else:
        reproject = reproject_chunk

    return reproject

def reproject_bulk(src_ds, example_ds, dst_path, spatial_interp = "nearest", **kwargs):
    """Reproject a dataset using `rioxarray.reproject_match`.

    Parameters
    ----------
    src_ds : xr.Dataset
        Dataset to be reprojected.
    example_ds : xr.Dataset
        Dataset to match to.
    dst_path : str
        Path to new file.
    spatial_interp : "nearest" | "bilinear" | "cubic" | "cubic_spline" | "lanczos" | "average" | "mode", optional
        Spatial interpolation to use, by default "nearest".

    Returns
    -------
    xr.Dataset
        Reprojected dataset.
    """

    resampling = {'nearest': 0,
                    'bilinear': 1,
                    'cubic': 2,
                    'cubic_spline': 3,
                    'lanczos': 4,
                    'average': 5,
                    'mode': 6}[spatial_interp]
    
    ds_match = src_ds.rio.reproject_match(example_ds, resampling = resampling)
    ds_match = ds_match.assign_coords({"x": example_ds.x, "y": example_ds.y})

    label = f"Applying `reproject_bulk` to {os.path.split(src_ds.encoding['source'])[-1]}:{list(src_ds.data_vars)[0]} ({spatial_interp})."
    
    ds_match = save_ds(ds_match, dst_path, encoding = "initiate", label = label)

    return ds_match

def reproject(src_ds, example_ds, dst_path, spatial_interp = "nearest", 
                max_bytes = 2e9, min_times = 10, stack_dim = "time"):
    """Reproject a dataset using `reproject_bulk` or `reproject_chunk`.

    Parameters
    ----------
    src_ds : xr.Dataset
        Dataset to be reprojected.
    example_ds : xr.Dataset
        Dataset to match to.
    dst_path : str
        Path to new file.
    spatial_interp : "nearest" | "bilinear" | "cubic" | "cubic_spline" | "lanczos" | "average" | "mode", optional
        Spatial interpolation to use, by default "nearest".
    max_bytes : int, optional
        Threshold value to switch to chunked reprojecting, by default 2e9.
    min_times : int, optional
        Threshold value to switch to bulk reprojecting, by default 10.
    stack_dim : str, optional
        Dimension name that is not spatial, by default "time".

    Returns
    -------
    xr.Dataset
        Reprojected dataset.
    """

    test_ds = [src_ds, example_ds][np.argmax([src_ds.nbytes, example_ds.nbytes])]

    reproj = choose_reprojecter(test_ds, max_bytes = max_bytes, min_times = min_times, stack_dim = stack_dim)

    if "source" in src_ds.encoding.keys():
        log.info(f"--> Selected `{reproj.__name__}` for reprojection of {os.path.split(src_ds.encoding['source'])[-1]}.").add()
    else:
        log.info(f"--> Selected `{reproj.__name__}` for reprojection.").add()

    ds = reproj(src_ds, example_ds, dst_path, spatial_interp = spatial_interp, stack_dim = stack_dim)
    
    log.sub()

    return ds

def reproject_chunk(src_ds, example_ds, dst_path, spatial_interp = "nearest", stack_dim = "time"):
    """Reproject a dataset using `rasterio.WarpedVRT`.

    Parameters
    ----------
    src_ds : xr.Dataset
        Dataset to be reprojected.
    example_ds : xr.Dataset
        Dataset to match to.
    dst_path : str
        Path to new file.
    spatial_interp : "nearest" | "bilinear" | "cubic" | "cubic_spline" | "lanczos" | "average" | "mode", optional
        Spatial interpolation to use, by default "nearest".

    Returns
    -------
    xr.Dataset
        Reprojected dataset.
    """

    das = list()
    variables = dict()
    ncs = list()

    if has_changed(src_ds) or not has_geotransform(src_ds):
        src_path = src_ds.encoding.get("source", dst_path).replace(".nc", "_fixed.nc")
        new_src_ds = src_ds.sortby("y", ascending = False)
        new_src_ds = new_src_ds.rio.write_transform(new_src_ds.rio.transform(recalc=True))
        src_ds = save_ds(new_src_ds, src_path, encoding = "initiate", label = "Correcting src_ds.")
        ncs.append(src_ds)
    else:
        src_path = src_ds.encoding["source"]

    resampling = {
                    'nearest': 0,
                    'bilinear': 1,
                    'cubic': 2,
                    'cubic_spline': 3,
                    'lanczos': 4,
                    'average': 5,
                    'mode': 6
                }[spatial_interp]

    example_ds = example_ds.sortby("y", ascending=False)

    vrt_options = {
        'resampling': resampling,
        'crs': example_ds.rio.crs,
        'transform': example_ds.rio.transform(recalc=True),
        'height': example_ds.y.size,
        'width': example_ds.x.size,
        'dtype': "float64",
    }

    for var in src_ds.data_vars:

        part_path = dst_path.replace(".nc", f"_{var}_temp.nc")

        @performance_check
        def _save_warped_vrt(src_path, var, vrt_options, part_path):
            with rasterio.open(f'NETCDF:{src_path}:{var}') as src:
                with WarpedVRT(src, **vrt_options) as vrt:
                    rio_shutil.copy(vrt, part_path, driver='netcdf')#, creation_options = {"COMPRESS": "DEFLATE"})

        _save_warped_vrt(src_path, var, vrt_options, part_path, label = "Warping VRT to netCDF.")

        ds_part = open_ds(part_path, decode_times = False)

        if stack_dim in src_ds.coords:
            da = ds_part.to_array(stack_dim, name = var).assign_coords({stack_dim: src_ds[stack_dim]})
        else:
            da = ds_part[var]

        das.append(da)
        ncs.append(ds_part)
        variables[var] = (None, var)

    ds = xr.merge(das)

    if "spatial_ref" not in ds.coords:
        ds = ds.assign_coords({"spatial_ref": ds.crs})

    ds = ds.drop_vars("crs")
    ds = process_ds(ds, {"x": ["lon", None], "y": ["lat", None]}, variables)
    ds = ds.assign_coords({"x": example_ds.x, "y": example_ds.y})

    label = f"Saving reprojected data from {os.path.split(src_ds.encoding['source'])[-1]}:{list(src_ds.data_vars)[0]} ({spatial_interp})."
    ds = save_ds(ds, dst_path, encoding = "initiate", label = label)

    for nc in ncs:
        remove_ds(nc)

    return ds

if __name__ == "__main__":



    src_ds = open_ds(r"/Users/hmcoerver/Local/custom_levels/MODIS/MOD13Q1.061.nc")
    example_ds = open_ds(r"/Users/hmcoerver/Local/custom_levels/LANDSAT/LC08_SR.nc")
    dst_path = r"/Users/hmcoerver/Local/custom_levels/test2.nc"

    # spatial_interp = "nearest"
#     var = "lst"

    ds = reproject_bulk(src_ds, example_ds, dst_path, spatial_interp = "nearest")



"""
https://docs.terrascope.be/#/Developers/WebServices/TerraCatalogue/ProductDownload
"""
import datetime
import requests
import os
import copy
import numpy as np
import xarray as xr
import pywapor.collect.protocol.cog as cog
import pywapor.general.bitmasks as bm
import pywapor.collect.accounts as accounts
from functools import partial
from osgeo import gdal
from joblib import Memory
import time
from requests.exceptions import ConnectionError
from pywapor.enhancers.apply_enhancers import apply_enhancers
from pywapor.general.logger import log, adjust_logger
from pywapor.general.processing_functions import save_ds, open_ds, remove_ds, adjust_timelim_dtype, process_ds, is_corrupt_or_empty

def default_post_processors(product_name, req_vars = ["ndvi", "r0"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary with a 
    list of functions per variable that should be applied after having collected the data
    from a server.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["ndvi", "r0"].

    Returns
    -------
    dict
        Functions per variable that should be applied to the variable.
    """

    post_processors = {
        "urn:eop:VITO:PROBAV_S5_TOC_100M_COG_V2": {
            "r0": [
                    calc_r0, 
                    partial(mask_bitwise_qa, 
                    flags = ["bad BLUE", "bad RED", "bad NIR", "bad SWIR", 
                            "sea", "undefined", "cloud", "ice/snow", "shadow"]),
            ],
            "ndvi": [
                    partial(mask_bitwise_qa, 
                    flags = ["bad RED", "bad NIR", "sea", "undefined", 
                            "cloud", "ice/snow", "shadow"]),
            ],
        }
    }

    out = {k:v for k,v in post_processors[product_name].items() if k in req_vars}

    return out

def default_vars(product_name, req_vars = ["ndvi", "r0"]):
    """Given a `product_name` and a list of requested variables, returns a dictionary
    with metadata on which exact layers need to be requested from the server, how they should
    be renamed, and how their dimensions are defined.

    Parameters
    ----------
    product_name : str
        Name of the product.
    req_vars : list, optional
        List of variables to be collected, by default ["ndvi", "r0"].

    Returns
    -------
    dict
        Metadata on which exact layers need to be requested from the server.
    """
    
    variables = {
        "urn:eop:VITO:PROBAV_S5_TOC_100M_COG_V2": {
            "NIR":      [("lon", "lat"), "nir"],
            "RED":      [("lon", "lat"), "red"],
            "BLUE":     [("lon", "lat"), "blue"],
            "SWIR":     [("lon", "lat"), "swir"],
            "SM":       [("lon", "lat"), "qa"],
            "GEOMETRY": [("lon", "lat"), "geometry"],
            "NDVI":     [("lon", "lat"), "ndvi"],
            "TIME":     [("lon", "lat"), "time"],
        },
    }

    req_dl_vars = {
        "urn:eop:VITO:PROBAV_S5_TOC_100M_COG_V2": {
            "ndvi": ["NDVI", "SM"],
            "r0":   ["BLUE", "NIR", "RED", "SWIR", "SM"],
        }
    }

    out = {val:variables[product_name][val] for sublist in map(req_dl_vars[product_name].get, req_vars) for val in sublist}
    
    return out

def mask_bitwise_qa(ds, var, flags):
    """Mask PROBAV data using a qa variable.

    Parameters
    ----------
    ds : xr.Dataset
        Input data.
    var : str
        Variable in `ds` to mask.
    flags : list
        Which flags not to mask.

    Returns
    -------
    xr.Dataset
        Masked dataset.
    """
    flag_bits = bm.PROBAV_qa_translator()
    mask = bm.get_mask(ds["qa"].astype("uint8"), flags, flag_bits)
    ds[var] = ds[var].where(~mask, np.nan)
    return ds

def calc_r0(ds, *args):
    ds["r0"] = 0.429 * ds["blue"] + 0.333 * ds["red"] + 0.133 * ds["nir"] + 0.105 * ds["swir"]
    return ds

def download(folder, latlim, lonlim, timelim, product_name, req_vars = ["ndvi", "r0"],
                variables = None, post_processors = None, timedelta = np.timedelta64(60, "h")):
    """Download MODIS data and store it in a single netCDF file.

    Parameters
    ----------
    folder : str
        Path to folder in which to store results.
    latlim : list
        Latitude limits of area of interest.
    lonlim : list
        Longitude limits of area of interest.
    timelim : list
        Period for which to prepare data.
    product_name : str
        Name of the product to download.
    req_vars : list, optional
        Which variables to download for the selected product, by default ["ndvi", "r0"].
    variables : dict, optional
        Metadata on which exact layers need to be requested from the server, by default None.
    post_processors : dict, optional
        Functions per variable that should be applied to the variable, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded data.
    """
    folder = os.path.join(folder, "TERRA")

    if not os.path.isdir(folder):
        os.makedirs(folder)

    fn = os.path.join(folder, f"{product_name.replace(':','_')}.nc")
    req_vars_orig = copy.deepcopy(req_vars)
    if os.path.isfile(fn):
        existing_ds = open_ds(fn)
        req_vars_new = list(set(req_vars).difference(set(existing_ds.data_vars)))
        if len(req_vars_new) > 0:
            req_vars = req_vars_new
            existing_ds = existing_ds.close()
        else:
            return existing_ds[req_vars_orig]
        
    spatial_buffer = True
    if spatial_buffer:
        dx = dy = 0.0033
        latlim = [latlim[0] - dy, latlim[1] + dy]
        lonlim = [lonlim[0] - dx, lonlim[1] + dx]

    timelim = adjust_timelim_dtype(timelim)

    if isinstance(variables, type(None)):
        variables = default_vars(product_name, req_vars)

    if isinstance(post_processors, type(None)):
        post_processors = default_post_processors(product_name, req_vars)
    else:
        default_processors = default_post_processors(product_name, req_vars)
        post_processors = {k: {True: default_processors[k], False: v}[v == "default"] for k,v in post_processors.items() if k in req_vars}

    bb = [lonlim[0],latlim[0],lonlim[1],latlim[1]]

    coords = {"x": ["lon", lonlim], "y": ["lat", latlim]}

    sd = datetime.datetime.strftime(timelim[0], "%Y-%m-%dT00:00:00Z")
    ed = datetime.datetime.strftime(timelim[1], "%Y-%m-%dT23:59:59Z")
    search_dates = f"{sd}/{ed}"
    params = dict()
    params['limit'] = 250
    params['bbox'] = bb
    params['datetime'] = search_dates
    params['collections'] = [product_name]

    products = search_stac(params, os.path.join(folder, "cache"))

    dss = list()

    log.info(f"--> Processing {len(products)} scenes.").add()

    un, pw = accounts.get("TERRA")

    gdal_config_options = {
        "GDAL_DISABLE_READDIR_ON_OPEN": "EMPTY_DIR",
    }

    warp_kwargs = dict()
    warp_kwargs["outputBounds"] = [lonlim[0], latlim[0], lonlim[1], latlim[1]]
    warp_kwargs["outputBoundsSRS"] = "epsg:4326"

    outs = list()

    for i, product in enumerate(products):

        out_fp = os.path.join(folder, f"{product['properties']['title']}.nc")

        log.info(f"--> ({i+1}/{len(products)}) Processing `{product['properties']['title']}`.").add()

        if os.path.isfile(out_fp):
            corrupt = is_corrupt_or_empty(out_fp)
            if corrupt:
                remove_ds(out_fp)
            else:
                out = open_ds(out_fp)
                outs.append(out)
        else:
            all_cog_urls = [v.get("href") for k, v in product["assets"].items()]
            cog_urls = [(key, all_cog_urls[np.argmax([key in x for x in all_cog_urls])]) for key in variables.keys()]
            dl_urls = [(x[0], x[1].replace("https://services", f"/vsicurl/https://{un}:{pw}@services")) for x in cog_urls]

            dss = list()
            for j, (var, url) in enumerate(dl_urls):
                var_file = out_fp.replace(".nc", f"_{var}_temp.nc")
                log.info(f"--> ({j+1}/{len(dl_urls)}) Downloading `{var}`.")

                if os.path.isfile(var_file):
                    corrupt = is_corrupt_or_empty(var_file)
                    if corrupt:
                        remove_ds(var_file)
                    else:
                        ds = open_ds(var_file)
                        dss.append(ds)
                else:
                    try:
                        for k, v in gdal_config_options.items():
                            gdal.SetConfigOption(k, v)
                        var_file = cog.cog_dl([url], var_file, warp_kwargs = warp_kwargs)
                    except Exception as e:
                        raise e
                    finally:
                        for k, v in gdal_config_options.items():
                            gdal.SetConfigOption(k, None)
                    ds = open_ds(var_file)
                    ds = ds.rename({"Band1": var})
                    dss.append(ds)
                    if var != dl_urls[-1][0]:
                        # NOTE VERY IMPORTANT, otherwise easily hitting "429" errors, 
                        # resulting in very strange behaviour.
                        time.sleep(10)

        ds = xr.merge(dss)
        ds = process_ds(ds, coords, variables)
        ds = ds.expand_dims({"time": 1}).assign_coords({"time": [datetime.datetime.strptime(product["properties"]["datetime"], "%Y-%m-%dT%H:%M:%SZ")]})

        # # Apply product specific functions.
        ds = apply_enhancers(post_processors, ds)

        out = save_ds(ds, out_fp, encoding = "initiate", label = f"Saving {os.path.split(out_fp)[-1]}.")
        outs.append(out)

        for x in dss:
            remove_ds(x)

        log.sub()
    
    log.sub()

    ds = xr.merge(outs)

    if not isinstance(timedelta, type(None)):
        ds["time"] = ds["time"] + timedelta

    ds = ds[req_vars]
    ds = ds.rio.write_crs("epsg:4326")

    ds = save_ds(ds, fn, label = f"Merging files.")

    for nc in outs:
        remove_ds(nc)

    return ds[req_vars_orig]

def search_stac(params, cachedir):

    memory = Memory(cachedir, verbose=0)

    search = 'https://services.terrascope.be/stac/search'

    @memory.cache()
    def _post_query(params):
        all_scenes = list()
        links = [{"body": params, "rel": "next"}]
        while "next" in [x.get("rel", None) for x in links]:
            params_ = [x.get("body") for x in links if x.get("rel") == "next"][0]
            try:
                query = requests.post(search, json = params_)
            except ConnectionError as e:
                log.info(f"--> The TERRA server is not responding (check `{search}`).")
                raise(e)
            query.raise_for_status()
            out = query.json()
            all_scenes += out["features"]
            links = out["links"]
        return all_scenes

    all_scenes = _post_query(params)

    return all_scenes
    
if __name__ == "__main__":

    variables = None
    post_processors = None

    folder = "/Users/hmcoerver/local/test_dl_TERRA_0"
    latlim = [29.4, 29.7]
    lonlim = [30.7, 31.0]
    timelim = [datetime.date(2020, 7, 2), datetime.date(2020, 7, 9)]
    product_name = "urn:eop:VITO:PROBAV_S5_TOC_100M_COG_V2"
    req_vars = ['r0']
    timedelta = np.timedelta64(60, "h")

    adjust_logger(True, folder, "INFO")





import os
import xarray as xr
from osgeo import gdal
import datetime
from pywapor.general.processing_functions import save_ds, remove_ds, open_ds
from pywapor.enhancers.dms.pyDMS import DecisionTreeSharpener
import xml.etree.ElementTree as ET
import pandas as pd
from pywapor.general.logger import log, adjust_logger
import matplotlib.pyplot as plt
import numpy as np
import pywapor
import re
from pywapor.general.performance import performance_check
from mpl_toolkits.axes_grid1 import make_axes_locatable

def gdal_stop_warnings_and_raise_errors():
    # https://gis.stackexchange.com/questions/43404/how-to-detect-a-gdal-ogr-warning/68042
    class GdalErrorHandler(object):
        def __init__(self):
            self.err_level=gdal.CE_None
            self.err_no=0
            self.err_msg=''
        def handler(self, err_level, err_no, err_msg):
            self.err_level=err_level
            self.err_no=err_no
            self.err_msg=err_msg
    err=GdalErrorHandler()
    handler=err.handler
    gdal.PushErrorHandler(handler)
    gdal.UseExceptions()

gdal_stop_warnings_and_raise_errors()

def highres_inputs(workdir, temporal_hr_input_data, static_hr_input_data, filter_times = None):
    """Create VRT files with all highres inputs per datetime.

    Parameters
    ----------
    workdir : str
        Folder in which to store the VRT files.
    temporal_hr_input_data : list
        Paths to the input highres data.
    static_hr_input_data : list
        Paths to static highres data.

    Returns
    -------
    list
        Paths to VRT files.
    """

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    # Find file with smallest pixels
    all_files = temporal_hr_input_data + static_hr_input_data
    template = all_files[np.argmin([abs(geot[1]) * abs(geot[5]) for fn, geot in [(x, gdal.Open(x).GetGeoTransform()) for x in all_files]])]

    # Determine bounds and resolution
    template_ds = gdal.Open(template)
    geot = template_ds.GetGeoTransform()
    xres = abs(geot[1])
    yres = abs(geot[5])
    bounds = (geot[0], geot[3] + template_ds.RasterYSize * geot[5], geot[0] + template_ds.RasterXSize * geot[1], geot[3])
    template_ds = None

    buildvrt_options = gdal.BuildVRTOptions(outputBounds = bounds, xRes = xres, yRes = yres, separate = True, bandList=[1])
    output_vrt = os.path.join(workdir, "highres_input_template.vrt")
    if os.path.isfile(output_vrt):
        try:
            os.remove(output_vrt)
        except PermissionError:
            log.info(f"--> Unable to delete temporary file `{output_vrt}`.")

    ds = gdal.BuildVRT(output_vrt, temporal_hr_input_data + static_hr_input_data, options = buildvrt_options)
    ds.FlushCache()
    ds = None

    ds = xr.open_dataset(re.findall(r"NETCDF:(.*):.*", temporal_hr_input_data[0])[0])
    times = [x.strftime("%Y%m%d_%H%M%S") for x in pd.to_datetime(ds.time.values)]
    if not isinstance(filter_times, type(None)):
        times = [x for x in times if x in filter_times]
    ds = ds.close()

    with open(output_vrt, 'r') as f:
        data = f.read()

    fps = list()
    for time_index, dt in enumerate(times):
        tree = ET.fromstring(data)
        for x in tree.findall("VRTRasterBand"):
            source_node = x.findall("ComplexSource")[0]
            fn_node = source_node.findall("SourceFilename")[0]
            if fn_node.text in temporal_hr_input_data:
                band_node = source_node.findall("SourceBand")[0]
                band_node.text = str(time_index + 1)

        fp_out = output_vrt.replace("_template.vrt", f"_{dt}.vrt")
        with open(fp_out, 'wb') as output:
            output.write(ET.tostring(tree))
        fps.append(fp_out)

    try:
        os.remove(output_vrt)
    except PermissionError:
        log.info(f"--> Unable to delete temporary file `{output_vrt}`.")

    return fps

def lowres_inputs(workdir, temporal_lr_input_data):
    """Create VRT files with all lowres inputs per datetime.

    Parameters
    ----------
    workdir : str
        Folder in which to store the VRT files.
    temporal_hr_input_data : list
        Paths to the input lowres data.

    Returns
    -------
    list
        Paths to VRT files.
    """

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    ds = xr.open_dataset(re.findall(r"NETCDF:(.*):.*", temporal_lr_input_data)[0], chunks = "auto")
    times = [x.strftime("%Y%m%d_%H%M%S") for x in pd.to_datetime(ds.time.values)]
    ds = ds.close()

    fps = list()
    for time_index, dt in enumerate(times):
        buildvrt_options = gdal.BuildVRTOptions(resolution = "highest", bandList = [time_index + 1])
        output_vrt = os.path.join(workdir, f"lowres_input_{dt}.vrt")
        if os.path.isfile(output_vrt):
            try:
                os.remove(output_vrt)
            except PermissionError:
                log.info(f"--> Unable to delete temporary file `{output_vrt}`.")
        ds = gdal.BuildVRT(output_vrt, [temporal_lr_input_data], options = buildvrt_options)
        ds.FlushCache()
        ds = None
        fps.append(output_vrt)

    return fps

def preprocess(ds):
    """Adds a time dimension to `ds`. Used internally by `sharpen` when calling `xr.open_mfdataset` to create
    a stacking dimension.

    Parameters
    ----------
    ds : xr.Dataset
        Input dataset.

    Returns
    -------
    xr.Dataset
        Dataset with added time dimension.
    """
    dt = datetime.datetime.strptime(os.path.split(ds.encoding["source"])[-1], "highres_output_%Y%m%d_%H%M%S.nc")
    ds["Band1"] = ds["Band1"].expand_dims({"time": 1}).assign_coords({"time": [dt]})
    return ds

def plot_sharpening(lr_fn, hr_fn, var, workdir):
    """Create a plot of a sharpened image.

    Parameters
    ----------
    lr_fn : str
        Path to lowres file.
    hr_fn : str
        Path to highres file.
    var : str
        Name of var.
    workdir : str
        Path to folder in which to save graphs.
    """

    if not os.path.exists(workdir):
        os.makedirs(workdir)

    fig, axs = plt.subplots(1,2, figsize=(10,6), dpi=300, sharex=True, sharey=True)
    
    # NOTE https://github.com/corteva/rioxarray/issues/580
    lr = xr.open_dataset(lr_fn).isel(band=0).band_data
    hr = xr.open_dataset(hr_fn).Band1
    lr_clipped = lr.rio.clip_box(*hr.rio.bounds()) * 0.0001

    lr_clipped.plot(ax = axs[0], add_colorbar=False, vmin = 280, vmax = 320)
    im1 = hr.plot(ax = axs[1], add_colorbar=False, vmin = 280, vmax = 320)

    divider = make_axes_locatable(axs[1])
    cax = divider.append_axes('right', size='5%', pad=0.05)
    fig.colorbar(im1, cax=cax, orientation='vertical', label = var)

    dt = datetime.datetime.strptime(os.path.split(hr.encoding["source"])[-1], "highres_output_%Y%m%d_%H%M%S.nc")
    fig.suptitle(dt.strftime("%Y-%m-%d %H:%M:%S"))
    axs[0].set_facecolor("lightgray")
    axs[1].set_facecolor("lightgray")
    axs[0].grid()
    axs[1].grid()
    axs[0].set_title(f"{lr.rio.resolution()[0]:.4f}° x {lr.rio.resolution()[1]:.4f}°")
    axs[1].set_title(f"{hr.rio.resolution()[0]:.4f}° x {hr.rio.resolution()[1]:.4f}°")
    axs[1].set_ylabel("")
    fig.savefig(os.path.join(workdir, dt.strftime("%Y%m%d_%H%M%S.jpg")))

    lr = lr.close()
    hr = hr.close()

def sharpen(dss, var, folder, *args, make_plots = False, req_vars = ['nmdi', 'bsi', 'mndwi',
                'vari_red_edge', 'psri', 'nir', 'green']):
    """Thermal sharpen datasets.

    Parameters
    ----------
    dss : dict
        Keys are variable names, values are xr.Datasets which contain the lowres and highres input data.
    var : str
        Variable name of the lowres input data (usually `lst`).
    folder : str
        Path to folder in which to store (intermediate) files.
    make_plots : bool, optional
        Whether or not to create plots, by default False.
    vars_for_sharpening : list, optional
        Variables to use as sharpening features, by default ['nmdi', 'bsi', 'mndwi', 'cos_solar_zangle', 'vari_red_edge', 'psri', 'nir', 'green', 'z', 'aspect', 'slope'].

    Returns
    -------
    dict
        Keys are variable names, values are (sharpened) datasets.
    """
    workdir = os.path.join(folder, "DMS")
    fp = os.path.join(workdir, f"{var}_i.nc")

    if not os.path.isfile(fp):
        # Open unopened netcdf files.
        dss = {**{k: open_ds(v) for k, v in dss.items() if isinstance(v, str)}, 
                **{k:v for k,v in dss.items() if not isinstance(v, str)}}

        if 'cos_solar_zangle' in req_vars and not 'cos_solar_zangle' in dss.keys():
            dss = get_cos_solar_zangle(dss, var, folder)
            remove_cos_solar_zangle = True
        else:
            remove_cos_solar_zangle = False

        out_fns = list()

        temporal_lr_input_data = f"NETCDF:{dss[var].encoding['source']}:{var}"
        temporal_hr_input_data = [f"NETCDF:{y.encoding['source']}:{x}" for x, y in dss.items() if "time" in y.dims and x in req_vars]
        static_hr_input_data = [f"NETCDF:{y.encoding['source']}:{x}" for x, y in dss.items() if "time" not in y.dims and x in req_vars]

        lowResFiles = sorted(lowres_inputs(workdir, temporal_lr_input_data))
        lowResDates = [os.path.split(x)[-1].split("_")[-2:] for x in lowResFiles]
        filter_times = [os.path.split(x)[-1].replace(".vrt", "").replace("lowres_input_", "") for x in lowResFiles]

        highResFiles = sorted(highres_inputs(workdir, temporal_hr_input_data, static_hr_input_data, filter_times = filter_times))
        highResDates = [os.path.split(x)[-1].split("_")[-2:] for x in highResFiles]
        
        assert np.all([x == y for x,y in zip(highResDates, lowResDates)])

        missing_vars = [x for x in req_vars if x not in dss.keys()]
        if len(missing_vars) > 0:
            log.info(f"--> Sharpening {len(highResFiles)} `{var}` images, without `{'` and `'.join(missing_vars)}` features.").add()
        else:
            log.info(f"--> Sharpening {len(highResFiles)} `{var}` images.").add()

        for i, (highResFilename, lowResFilename) in enumerate(zip(highResFiles, lowResFiles)):
            fp_, fn = os.path.split(highResFilename)

            lbl = f"({i+1}/{len(highResFiles)}) Sharpening `{os.path.split(lowResFilename)[-1]}` with `{fn}`."

            _, correctedImage = thermal_sharpen(highResFilename, lowResFilename, label = lbl)

            fp_out = os.path.join(fp_, fn.replace("input", "output").replace(".vrt", ".nc"))
            ds = gdal.Translate(fp_out, correctedImage)
            ds.FlushCache()
            ds = None

            if make_plots:
                plot_sharpening(lowResFilename, fp_out, var, workdir)

            out_fns.append(fp_out)

            try:
                os.remove(highResFilename)
            except PermissionError:
                log.info(f"--> Unable to delete temporary file `{highResFilename}`.")

            try:
                os.remove(lowResFilename)
            except PermissionError:
                log.info(f"--> Unable to delete temporary file `{lowResFilename}`.")

        log.sub()

        dss_ = [preprocess(open_ds(x)) for x in out_fns]
        ds_ = xr.concat(dss_, dim = "time").sortby("time")
        ds_ = ds_.rename_dims({"lat": "y", "lon": "x"})
        ds_ = ds_.transpose("time", "y", "x")
        ds_ = ds_.rename_vars({"crs": "spatial_ref", "Band1": var, "lat": "y", "lon": "x"})
        ds_ = ds_.rio.write_grid_mapping("spatial_ref")
        ds_ = ds_.sortby("y", ascending = False)
        ds_ = ds_.sortby("x")
        ds_.attrs = {}

        ds = save_ds(ds_, fp, encoding = "initiate", label = f"Merging sharpened `{var}` files.")

        ds_ = ds_.close()

        for x in dss_:
            remove_ds(x)
        
        if 'cos_solar_zangle' in dss.keys() and remove_cos_solar_zangle:
            remove_ds(dss['cos_solar_zangle'])
            _ = dss.pop("cos_solar_zangle")

    dss[var] = fp

    return dss

@performance_check
def thermal_sharpen(highres_fn, lowres_fn):
    """Sharpen a single lowres file using a single highres file (in which each band contains one feature).

    Parameters
    ----------
    highres_fn : str
        Path to highres input.
    lowres_fn : str
        Path to lowres input.

    Returns
    -------
    tuple
        `gdal.Dataset`s of the residual and the corrected image.
    """

    commonOpts = {
                "highResFiles":                     [highres_fn],
                "lowResFiles":                      [lowres_fn],
                "lowResQualityFiles":               [],
                "lowResGoodQualityFlags":           [],
                "cvHomogeneityThreshold":           0.0,
                "movingWindowSize":                 0,
                "disaggregatingTemperature":        True,
            }

    dtOpts = {
            "perLeafLinearRegression":              True,
            "linearRegressionExtrapolationRatio":   0.25
            }

    opts = commonOpts.copy()
    opts.update(dtOpts)
    disaggregator = DecisionTreeSharpener(**opts)

    disaggregator.trainSharpener()
    downscaledFile = disaggregator.applySharpener(highres_fn, lowres_fn)
    residualImage, correctedImage = disaggregator.residualAnalysis(downscaledFile, lowres_fn, lowResQualityFilename = True)

    return residualImage, correctedImage

def get_cos_solar_zangle(dss, var, folder):
    """Calculate the cosine solar zenith angle.

    Parameters
    ----------
    dss : dict
        Keys are variables, values are datasets.
    var : str
        Variable from which the time is used to calculate `cos_solar_zangle` at.
    folder : str
        Path to folder to store files.

    Returns
    -------
    dict
        Keys are variables, values are datasets.
    """
    reqs = ["slope", "aspect", var]
    checks = [(x in dss.keys()) and (x in dss[x].data_vars) for x in reqs]
    if np.all(checks):
        lat_deg = xr.DataArray(dss["slope"]["y"].values, coords = {"y":dss["slope"]["y"]}).rename("lat").chunk("auto")
        lat = pywapor.et_look_v2_v3.solar_radiation.latitude_rad(lat_deg)
        slope = dss["slope"]["slope"]
        aspect = dss["aspect"]["aspect"]
        doy = dss[var]["time"].dt.strftime("%j").astype(int).chunk("auto")
        dtime = (dss[var]["time"].dt.strftime("%H").astype(int) + dss[var]["time"].dt.strftime("%M").astype(int)/60.).chunk("auto")
        sc = pywapor.et_look_v2_v3.solar_radiation.seasonal_correction(doy)
        decl = pywapor.et_look_v2_v3.solar_radiation.declination(doy)
        ha = pywapor.et_look_v2_v3.solar_radiation.hour_angle(sc, dtime)
        ds = pywapor.et_look_v2_v3.solar_radiation.cosine_solar_zenith_angle(ha, decl, lat, slope, aspect).rename('cos_solar_zangle').to_dataset()
        ds = ds.rio.write_crs(aspect.rio.crs)
        ds = save_ds(ds, os.path.join(folder, "cos_solar_zangle_i.nc"), label = "Calculating `cos_solar_zangle`.")
        dss['cos_solar_zangle'] = ds
    else:
        log.warning(f"--> Couldn't calculate `cos_solar_zangle`, `{'` and `'.join([x for x, check in zip(reqs,checks) if not check])}` is missing.")

    return dss

if __name__ == "__main__":

    folder = "/Users/hmcoerver/Local/pywapor_se"

    adjust_logger(True, folder, "INFO")
    make_plots = False

    # vars_for_sharpening = ['ndvi', 'bsi', 'mndwi', 'green', 'z']

    # var = "bt"
    # ds = open_ds(r"/Users/hmcoerver/Local/pywapor_se/VIIRSL1/VNP02IMG.nc")

    # dss = {
    #     'bt': ds,
    #     'z': open_ds('/Users/hmcoerver/Local/pywapor_se/z_i.nc'),
    #     'ndvi':             open_ds('/Users/hmcoerver/Local/pywapor_se/ndvi_i.nc'),
    #     'bsi':              open_ds('/Users/hmcoerver/Local/pywapor_se/bsi_i.nc'),
    #     'mndwi':            open_ds('/Users/hmcoerver/Local/pywapor_se/mndwi_i.nc'),
    #     'green':            open_ds('/Users/hmcoerver/Local/pywapor_se/green_i.nc'),
    #     }

    # highres_fn = "/Users/hmcoerver/Local/pywapor_se/DMS/highres_input_20220108_061800.vrt"
    # lowres_fn = "/Users/hmcoerver/Local/pywapor_se/DMS/lowres_input_20220108_061800.vrt"
    import matplotlib.pyplot as plt

    highres_fn = "/Users/hmcoerver/Local/pywapor_se/DMS/highres_input_20220108_080000.vrt"
    lowres_fn = "/Users/hmcoerver/Local/pywapor_se/DMS/lowres_input_20220108_080000.vrt"


    # out = sharpen(dss, var, folder)


"""
Functions to automatically decorate all the functions inside the et_look_v2 and
et_look_dev submodules so that et_look can be run when not all inputs variables 
are available. This allows to run only parts of et_look, e.g. when only `et_ref_24_mm`
as output is required, pre_et_look doesn't need to download and process all variables.
"""
import types
import numpy as np
from pywapor.general.logger import log

def decorate_mod(module, decorator):
    """Apply a decorator to all the functions inside a module.

    Parameters
    ----------
    module : module
        Module to decorate.
    decorator : function
        Function to decorate with.
    """
    for name in dir(module):
        if name not in [
                        "ra_soil",
                        "initial_friction_velocity_inst",
                        "monin_obukhov_length", 
                        "stability_parameter", 
                        "stability_factor", 
                        "friction_velocity", 
                        "ra_canopy",
                        "calc_dlat_dlon",
                        "calc_slope",
                        "stability_parameter_obs", 
                        "stability_correction_heat_obs"]:
            obj = getattr(module, name)
            if isinstance(obj, types.FunctionType) and not hasattr(obj, 'decorated'):
                setattr(module, name, decorator(obj))

def decorate_submods(module, decorator):
    """Apply a decorator to all the functions inside all the submodules of a 
    module.

    Parameters
    ----------
    module : module
        Module of which the functions inside its submodules to decorate.
    decorator : function
        Function to decorate with.
    """
    for submod in dir(module):
        submod = getattr(module, submod)
        if isinstance(submod, types.ModuleType):
            decorate_mod(submod, decorator)

def etlook_decorator(func):
    """Checks if the DataArrays contain data or are None.

    Parameters
    ----------
    func : function
        Function for which to check the inputs.
    """
    group = str(func.__module__)
    def wrapper_func(*args, **kwargs):
        check1 = np.all([arg.dtype != object for arg in args])
        check2 = np.all([arg.dtype != object for _, arg in kwargs.items()])
        if check1 and check2:
            log.info(f"--> Calculating `{func.__name__}`.")
            x = func(*args, **kwargs)
            x.attrs["calculated_with"] = [arg.name for arg in args]
            x.attrs["et_look_module"] = group
            return x
        else:
            missing_args = [arg.name for arg in args if arg.dtype == object]
            missing_kwargs = [name for name, arg in kwargs.items() if arg.dtype == object]
            missing = missing_args + missing_kwargs
            log.warning(f"--> Insufficient data found for `{func.__name__}`, missing [`{'`, `'.join(missing)}`]")
    wrapper_func.__module__ = func.__module__
    wrapper_func.__name__ = func.__name__
    setattr(wrapper_func, "decorated", True)
    return wrapper_func


import pywapor
import os
import numpy as np
import glob
import warnings
import json
from functools import partial
from pywapor.general.logger import log, adjust_logger
from pywapor.collect.downloader import collect_sources
from pywapor.general.levels import pre_et_look_levels
from pywapor.general.compositer import time_bins
from pywapor.general.processing_functions import adjust_timelim_dtype, func_from_string, open_ds

class Configuration():

    variable_categories = {
        "optical": ["ndvi", "r0", "mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"],
        "thermal": ["bt", "lst"],
        "solar radiation": ["ra"],
        "precipitation": ["p"],
        "elevation": ["z", "slope", "aspect"],
        "meteorological": ["t_air", "t_air_min", "t_air_max", "u", "vp", "p_air", "p_air_0", "u2m", "v2m", "qv", "wv", "t_dew"],
        "statics": ["lw_slope", "lw_offset", "z_obst_max", "rs_min", "land_mask", "vpd_slope", "t_opt", "t_amp", "t_amp_year", "rn_slope", "rn_offset", "z_oro"],
        "soil moisture": ["se_root"],
    }

    category_variables = {}
    for x in [{var: cat for var in varis} for cat, varis in variable_categories.items()]:
        category_variables = {**category_variables, **x}

    se_root_vars = [
        [("ndvi",)], 
        [("bt", "lst"), ("lst",), ("bt",)], 
        [("u",), ("u2m", "v2m")],
        [("p_air",)], 
        [("p_air_0",)], 
        [("t_air",)], 
        [("wv",)], 
        [("t_dew",), ("qv",)]
        ]
    
    et_look_vars = [
        [("ndvi",)],
        [("r0",)],

        [("se_root",)], 

        [("p",)],
        [("z",)],
        [("ra",)],

        [("u",), ("u2m", "v2m")],
        [("p_air",)], 
        [("p_air_0",)], 
        [("t_air",)],
        [("t_air_min",)], 
        [("t_air_max",)], 
        [("t_dew",), ("qv",)],

        [("rn_offset",)],
        [("rn_slope",)],
        [("t_amp",),],
        [("t_opt",)],
        [("vpd_slope",)],
        [("lw_offset",)],
        [("lw_slope",)],

        [("rs_min",), ()],
        [("land_mask",), ()],
        [("z_obst_max",), ()],
        [("z_oro",), ()],
        ]
    
    @staticmethod
    def source_func(x):
        return x if "FILE:" in x else x.split(".")[0]
    
    @staticmethod
    def pname_func(x):
        return "from_file" if "FILE:" in x else ".".join(x.split(".")[1:])

    def __init__(self, full = None, summary = None, se_root = None, 
                 et_look = None):
        self.full = full
        self.summary = summary
        self.se_root = se_root
        self.et_look = et_look

    def __repr__(self):
        summary = self.summary.copy()

        example_product = summary.pop('_EXAMPLE_')
        whittaker = summary.pop('_WHITTAKER_')
        sharpen = summary.pop('_ENHANCE_')

        base = "\n".join([f"--> {cat.title()} data:\n    > `{'`, `'.join(sources)}`" for cat, sources in summary.items()])

        additional = f"""
--> Example Product:
    > {example_product}
--> Applying whittaker to variables from sources:
    > `{'` and `'.join(whittaker)}`
--> Sharpening these variables:
    > `{'` and `'.join(sharpen)}`"""
        return base + additional

    @staticmethod
    def default_enhancers(prod, var):
        source = Configuration.source_func(prod)
        if "FILE:" in source:
            return []
        product_name = Configuration.pname_func(prod)
        mod = getattr(pywapor.collect.product, source)
        x = mod.default_post_processors(product_name, [var])[var]

        funcs = list()
        for x_ in x:
            if isinstance(x_, partial):
                funcs.append({
                    "func": f"{x_.func.__module__}.{x_.func.__name__}",
                    "args": x_.args,
                    "keywords": x_.keywords,
                })
            else:
                funcs.append({"func": f"{x_.__module__}.{x_.__name__}"})
        
        return funcs

    @classmethod
    def from_name(cls, name):
        log.info(f"--> Searching configuration for `{name}`.").add()

        folder = os.path.realpath(pywapor.__path__[0])
        available_configs = glob.glob(os.path.join(folder, "configs", "*.json"))
        available_names = sorted([os.path.splitext(os.path.split(x)[-1])[0] for x in available_configs])
        log.info(f"> Available configurations are `{'`, `'.join(available_names)}`")
    
        synonyms = {
            "level_1": "WaPOR2_level_1",
            "level_2": "WaPOR2_level_2",
            "level_3": "WaPOR2_level_3",
            "level_1_v2": "WaPOR2_level_1",
            "level_2_v2": "WaPOR2_level_2",
            "level_3_v2": "WaPOR2_level_3",
            "level_1_v3": "WaPOR3_level_1",
            "level_2_v3": "WaPOR3_level_2",
            "level_3_v3": "WaPOR3_level_3",
        }

        name_ = synonyms.get(name, name)

        selection = None
        for fh in available_configs:
            fn = os.path.splitext(os.path.split(fh)[-1])[0]
            if fn == name_:
                selection = fh
                break

        if isinstance(selection, type(None)):
            raise ValueError(f"No configuration found for name `{name}`, choose one from `{'`, `'.join(available_names)}`.")
        else:
            config = Configuration.from_json(selection)

        log.sub().info("--> Configuration set.")

        return config
    
    def to_json(self, fh):
        out = {
            "summary": self.summary, 
            "full": self.full, 
            "se_root": self.se_root, 
            "et_look": self.et_look,
        }
        class EncodeSetTuple(json.JSONEncoder):
            def encode(self, obj):
                def hint_tuples(item):
                    if isinstance(item, tuple):
                        return {'type': "tuple", 'value': list(item)}
                    if isinstance(item, list):
                        return [hint_tuples(e) for e in item]
                    if isinstance(item, set):
                        return {'type': "set", 'value': list(item)}
                    if isinstance(item, dict):
                        return {key: hint_tuples(value) for key, value in item.items()}
                    else:
                        return item
                return super(EncodeSetTuple, self).encode(hint_tuples(obj))
        
        json_string = json.dumps(out, cls = EncodeSetTuple)
        with open(fh, "w") as x:
            x.write(json_string)

    @classmethod
    def from_json(cls, fh):
        log.info(f"--> Loading configuration from `{os.path.split(fh)[1]}`.").add()
        def decode_set(dct):
            if dct.get("type", None) == "set":
                return set(dct.get('value'))
            if dct.get("type", None) == "tuple":
                return tuple(dct.get('value'))
            return dct
        with open(fh, "r") as fp:
            config_ = json.load(fp, object_hook=decode_set)
        
        config = cls(full = config_["full"], summary = config_["summary"], se_root = config_["se_root"], 
                 et_look = config_["et_look"])

        config.validate()
        log.sub().info("--> Configuration loaded.")
        return config
    
    @classmethod
    def from_summary(cls, summary):
        log.info(f"--> Creating configuration from summary.").add()
        example_product = summary.pop('_EXAMPLE_', '')
        temporal_interp = summary.pop('_TEMP_INTERP_', {})
        enhance = summary.pop('_ENHANCE_', {})

        full = dict()

        unique_enhancers = set([item for row in list(enhance.values()) for item in row])
        extra_vars = list()
        for enhancer in unique_enhancers:
            func = func_from_string(enhancer)
            extra_vars += [[(x,)] for x in func.__kwdefaults__["req_vars"]]

        for var_group in cls.et_look_vars + cls.se_root_vars + extra_vars:
            for var_ in var_group:
                for var in var_:
                    cat = cls.category_variables[var]
                    possible_prods = summary[cat]
                    products = []
                    t_interps = list()
                    for prod in possible_prods:
                        valid = cls.has_var(prod, var)
                        if valid:

                            enhancers = cls.default_enhancers(prod, var)

                            products += [{
                                "source": cls.source_func(prod), 
                                "product_name": cls.pname_func(prod), 
                                "enhancers": enhancers,
                                "is_example": prod == example_product,
                                }]
                        t_interps += [temporal_interp.get(prod, [])]
                    
                    if len(products) == 0:
                        continue

                    if len(t_interps) == 0:
                        temporal_interp_ = "linear"
                    elif len(t_interps) == 1:
                        temporal_interp_ = t_interps[0]
                    else:
                        log.warning("")
                        temporal_interp_ = t_interps[0]

                    variable_enhancers = enhance.get(var, [])

                    composite_type = "mean" if var.split("_")[-1] not in ["min", "max"] else var.split("_")[-1]

                    full[var] = {
                        "products": products,
                        "temporal_interp": temporal_interp_,
                        "variable_enhancers": variable_enhancers,
                        "spatial_interp": "bilinear",
                        "composite_type": composite_type,
                    }

        config = cls(full = full)
        config.validate()
        config.summarize()
        config.update_se_root_config()
        config.update_et_look_config()
        log.sub().info("--> Configuration created.")
        return config

    @staticmethod
    def has_var(prod, var, verbose = True):
        source = Configuration.source_func(prod)
        product_name = Configuration.pname_func(prod)
        if "FILE:" in source:
            if not verbose:
                log.info(f"> Variable `{var}` will be loaded from a file `{source}`.")
            return True
        mod = getattr(pywapor.collect.product, source)
        try:
            mod.default_vars(product_name, [var])
            valid = True
        except TypeError:
            valid = False
            if not verbose:
                log.warning(f"> {source}.{product_name} does not have a variable called `{var}`.") 
        return valid
    
    def validate(self):
        log.info("--> Validating configuration.").add()
        valids = []
        for var in self.full:
            for product in self.full[var]["products"]:
                prod = f"{product['source']}.{product['product_name']}"
                valid = Configuration.has_var(prod, var, verbose = False)
                valids.append(valid)
        if all(valids):
            log.info("> All specified variables sources are valid.")
        log.sub()
        return valid

    def update_se_root_config(self):
        log.info("--> Making configuration for SE_ROOT.").add()
        
        sharpened_vars = {var for var, config in self.full.items() if any(["sharpen" in x for x in config["variable_enhancers"]])}

        sharpeners = ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]
    
        se_root_config = dict()
        for var_group in self.se_root_vars:
            for var in var_group:
                if all([var_ in self.full.keys() for var_ in var]):
                    for var_ in var:
                        se_root_config[var_] = self.full[var_].copy()
                        _ = se_root_config[var_].pop("composite_type")
                    break
                elif var == var_group[-1]:
                    log.warning(f"--> No configuration found for {' or '.join([f"[`{'` and `'.join(x)}`]" for x in var_group])}.")
                else:
                    continue
        
        to_sharpen = set(sharpened_vars) & set(se_root_config)
        if to_sharpen:
            for var in sharpeners:
                if var in self.full.keys():
                    se_root_config[var] = self.full[var]
                else:
                    log.warning(f"--> No configuration found for `{var}` (required for sharpening of `{'` and `'.join(to_sharpen)}`).")

        self.se_root = se_root_config
        log.sub()

    def update_et_look_config(self):
        log.info("--> Making configuration for ET_LOOK.").add()
        
        sharpened_vars = {var for var, config in self.full.items() if any(["sharpen" in x for x in config["variable_enhancers"]])}

        sharpeners = ["mndwi", "psri", "vari_red_edge", "bsi", "nmdi", "green", "nir"]
    
        et_look_config = dict()
        for var_group in self.et_look_vars:
            for var in var_group:
                if all([var_ in self.full.keys() for var_ in var]):
                    for var_ in var:
                        et_look_config[var_] = self.full[var_].copy()
                    if len(var) == 0 and var == var_group[-1]:
                        log.info(f"--> No configuration found for {' or '.join([f"[`{'` and `'.join(x)}`]" for x in var_group[:-1]])}, will use constant value.")
                    break
                elif var == var_group[-1]:
                    log.warning(f"--> No configuration found for {' or '.join([f"[`{'` and `'.join(x)}`]" for x in var_group])}.")
                else:
                    continue
        
        to_sharpen = set(sharpened_vars) & set(et_look_config)
        if to_sharpen:
            for var in sharpeners:
                if var in self.full.keys():
                    et_look_config[var] = self.full[var]
                else:
                    log.warning(f"--> No configuration found for `{var}` (required for sharpening of `{'` and `'.join(to_sharpen)}`).")

        self.et_look = et_look_config
        log.sub()

    def summarize(self):
        log.info("--> Making summary of configuration.")
        
        summary = dict()
        for var, config in self.full.items():
            sources = set([f'{x["source"]}.{x["product_name"]}' for x in config["products"]])
            cat = self.category_variables.get(var, "unknown")
            if cat == "other":
                log.warning(f"--> Unknown variable `{var}` found.")
            if cat in list(summary.keys()):
                summary[cat].union(sources)
            else:
                summary[cat] = sources

        whittaker = dict()
        sharpen = set()
        for k,v in self.full.items():
            example_ = [f"{prod["source"]}.{prod["product_name"]}" for prod in v["products"] if prod["is_example"]]
            if len(example_) > 0:
                example = example_[0]
            if isinstance(v["temporal_interp"], dict):
                if v["temporal_interp"].get("method", "") == "whittaker":
                    for prod in v["products"]:
                        whittaker[f"{prod["source"]}.{prod["product_name"]}"] = v["temporal_interp"]
            if any(["sharpen" in x for x in v["variable_enhancers"]]):
                sharpen.add(k)

        summary["_WHITTAKER_"] = whittaker
        summary["_ENHANCE_"] = sharpen
        summary["_EXAMPLE_"] = example

        self.summary = summary

class Project():

    def __init__(self, project_folder, bb, period, configuration = None):

        assert bb[::2][0] < bb[::2][1], "Invalid Bounding-Box"
        self.lonlim = bb[::2]
        assert bb[1::2][0] < bb[1::2][1], "Invalid Bounding-Box"
        self.latlim = bb[1::2]
        if not os.path.isdir(project_folder):
            os.makedirs(project_folder)
        self.folder = project_folder
        self.period = adjust_timelim_dtype(period)
        self.configuration = configuration
        self.dss = None

        warnings.filterwarnings("ignore", message="invalid value encountered in power")

        adjust_logger(True, self.folder, "INFO")

        if os.path.isfile(os.path.join(self.folder, "se_root_in.nc")):
            self.se_root_in = open_ds(os.path.join(self.folder, "se_root_in.nc"))
        else:
            self.se_root_in = None

        se_root_outs = glob.glob(os.path.join(self.folder, "se_root_out*.nc"))
        if se_root_outs:
            self.se_root_out = open_ds(max(se_root_outs, key=os.path.getmtime))
        else:
            self.se_root_out = None

        if os.path.isfile(os.path.join(self.folder, "et_look_in.nc")):
            self.et_look_in = open_ds(os.path.join(self.folder, "et_look_in.nc"))
        else:
            self.et_look_in = None

        et_look_outs = glob.glob(os.path.join(self.folder, "et_look_out*.nc"))
        if et_look_outs:
            self.et_look_out = open_ds(max(et_look_outs, key=os.path.getmtime))
        else:
            self.et_look_out = None

        log.info("> PROJECT").add()
        log.info(self.__repr__())
        log.sub().info("< PROJECT")

    def __repr__(self):
        project_str = f"""--> Project Folder:
        > {self.folder}
    --> Period:
        > {self.period[0]} - {self.period[1]}
    --> Bounding-Box:\n
                 {self.latlim[1]:8.4f}
                 ┌─────────┐
                 │         │
       {self.lonlim[0]:9.4f} │         │{self.lonlim[1]:9.4f}
                 │         │
                 └─────────┘
                 {self.latlim[0]:8.4f}\n
    --> Configuration:
        > {self.configuration}
    --> pyWaPOR Version:
        > {pywapor.__version__}"""
        return project_str

    def load_configuration(self, name = None, summary = None, json = None):
        if not isinstance(name, type(None)) and not isinstance(summary, type(None)):
            raise ValueError("Only one of `level` and `summary` can be specified.")
        log.info("> CONFIGURATION").add()
        if not isinstance(name, type(None)):
            self.configuration = Configuration.from_name(name)
        elif not isinstance(summary, type(None)):
            summary = summary.copy()
            self.configuration = Configuration.from_summary(summary)
        elif not isinstance(json, type(None)):
            self.configuration = Configuration.from_json(json)
        else:
            raise ValueError("At least one of `level` or `summary` needs to be specified.")
        
        log.info("--> Summary:").add()
        summary_str = self.configuration.__repr__()

        for line in summary_str.split("\n"):
            log.info(line)

        log.sub().sub().info("< CONFIGURATION")
        return self.configuration
    
    def set_passwords(self):

        log.info("> PASSWORDS").add()

        req_accounts = {
            "MODIS": "NASA",
            "SRTM": "NASA",
            "CHIRPS": "NASA",
            "MERRA2": "NASA",
            "TERRA": "TERRA",
            "ERA5": "ECMWF",
            "LANDSAT": "EARTHEXPLORER",
            "SENTINEL2": "COPERNICUS_DATA_SPACE",
            "SENTINEL3": "COPERNICUS_DATA_SPACE",
        }

        all_accounts = list()
        for v in self.configuration.summary.values():
            if isinstance(v, set):
                for source_product in v:
                    name = Configuration.source_func(source_product)
                    account = req_accounts.get(name, None)
                    if not isinstance(account, type(None)):
                        all_accounts.append(account)

        x = set(all_accounts)
        log.info(f"--> Accounts needed for `{'`, `'.join(x)}`.").add()

        for account in x:
            _ = pywapor.collect.accounts.get(account)

        log.sub().info("--> All set!")
        log.sub().info("< PASSWORDS")

    def download_data(self, buffer_timelim = True):
        assert not isinstance(self.configuration, type(None)), "Please load a configuration before continueing."

        log.info("> DOWNLOADER").add()

        if buffer_timelim:
            bins = time_bins(self.period, 1)
            adjusted_timelim = [bins[0], bins[-1]]
            buffered_timelim = [adjusted_timelim[0] - np.timedelta64(3, "D"), 
                                adjusted_timelim[1] + np.timedelta64(3, "D")]
        else:
            adjusted_timelim = self.period
            buffered_timelim = self.period

        self.dss, _ = collect_sources(
                            self.folder,
                            self.configuration.full,
                            self.latlim,
                            self.lonlim,
                            buffered_timelim,
                            )
        
        log.sub().info("< DOWNLOADER")
    
        return self.dss
    
    def run_pre_se_root(self, forced = False):
        if isinstance(self.se_root_in, type(None)) or forced:
            self.se_root_in = pywapor.pre_se_root.main(self.folder, self.latlim, self.lonlim, self.period, sources = self.configuration.se_root)
        else:
            log.info("> PRE_SE_ROOT").add()
            log.info(f"--> Re-using `{os.path.split(self.se_root_in.encoding['source'])[-1]}`.")
            log.sub().info("< PRE_SE_ROOT")
        return self.se_root_in

    def run_se_root(self, se_root_version = "v3"):
        self.se_root_out = pywapor.se_root.main(self.se_root_in, se_root_version = se_root_version)
        return self.se_root_out
    
    def run_pre_et_look(self, forced = False):
        if isinstance(self.et_look_in, type(None)) or forced:
            self.et_look_in = pywapor.pre_et_look.main(self.folder, self.latlim, self.lonlim, self.period, sources = self.configuration.et_look)
        else:
            log.info("> PRE_ET_LOOK").add()
            log.info(f"--> Re-using `{os.path.split(self.et_look_in.encoding['source'])[-1]}`.")
            log.sub().info("< PRE_ET_LOOK")
        return self.et_look_in

    def run_et_look(self, et_look_version = "v3"):
        self.et_look_out = pywapor.et_look.main(self.et_look_in, et_look_version = et_look_version)
        return self.et_look_out

if __name__ == "__main__":

    timelim = ["2022-01-01", "2023-12-31"]
    latlim = [21.9692194682626933, 21.9939120838340507]
    lonlim = [91.9371349243682801, 91.9657566608824339]
    bb = [91.9371349243682801, 21.9692194682626933, 91.9657566608824339, 21.9939120838340507]
    project_folder = r"/Users/hmcoerver/Local/pywapor_bgd"

    adjust_logger(True, project_folder, "INFO")

    all_configs = {
        "WaPOR2_level_1",
        "WaPOR2_level_2",
        "WaPOR2_level_3",
        "WaPOR2_level_1",
        "WaPOR2_level_2",
        "WaPOR2_level_3",
        "WaPOR3_level_2",
        "WaPOR3_level_3",
        "nrt",
        "all_in",
    }

    project = pywapor.Project(project_folder, bb, timelim)

    # project.load_configuration(name = "WaPOR3_level_2")

    summary = {
        # Define which products to use.
        'elevation': {'COPERNICUS.GLO30'},
        'meteorological': {'GEOS5.tavg1_2d_slv_Nx'},
        'optical': {'SENTINEL2.S2MSI2A_R20m'},
        'precipitation': {'CHIRPS.P05'},
        'solar radiation': {'ERA5.sis-agrometeorological-indicators'},
        'statics': {'STATICS.WaPOR3'},
        'thermal': {'VIIRSL1.VNP02IMG'},
        'soil moisture': {'FILE:{folder}{sep}se_root_out*.nc.from_file'},

        # Define which product to reproject the other products to.
        '_EXAMPLE_': 'SENTINEL2.S2MSI2A_R20m', 

        # Define any special functions to apply to a specific variable.
        '_ENHANCE_': {"bt": ["pywapor.enhancers.dms.thermal_sharpener.sharpen"],},

        # Choose which products should be gapfilled.
        '_WHITTAKER_': {
            'SENTINEL2.S2MSI2A_R20m': {'lmbdas': 1000.0, 'method': 'whittaker'}, 
            'VIIRSL1.VNP02IMG': {'a': 0.85, 'lmbdas': 1000.0, 'method': 'whittaker'}},
        }
    
    project.load_configuration(summary = summary)



    project.set_passwords()
    dss = project.download_data()
    
    se_root_in = project.run_pre_se_root()
    # se_root = project.run_se_root()
    # et_look_in = project.run_pre_et_look()
    # et_look = project.run_et_look()








import numpy as np
from osgeo import gdal
import pywapor.general.processing_functions as PF
import warnings
import os
import datetime
try:
    import IPython
    valids = ['inline', 'osx']
    for valid in valids:
        # print(f"Trying {valid}")
        try:
            IPython.get_ipython().run_line_magic("matplotlib", valid)
            # print(f"Using {valid}")
            break
        except:
            continue
except:
    print("")
import matplotlib.pyplot as plt
from matplotlib.patches import Patch

warnings.filterwarnings("ignore", message="Mean of empty slice")

def prepare_dss(dss, sorter):
    for ds in dss:
        if "offset" not in ds.keys():
            ds["offset"] = None
        if "scale" not in ds.keys():
            ds["scale"] = None
        if "masks" not in ds.keys():
            ds["masks"] = None

    for ds in dss:
        to_sort = list(ds.keys())
        to_sort.append(to_sort.pop(to_sort.index(sorter)))
        for key in to_sort:
            if isinstance(ds[key], np.ndarray):
                ds[key] = ds[key].tolist()
            if not isinstance(ds[key], list):
                ds[key] = [ds[key]] * len(ds[sorter])
            ds[key] = [x for _, x in sorted(zip(ds[sorter], ds[key]))]

    for ds in dss:
        assert dss[0][sorter] == ds[sorter]

    return dss

def determine_offsets(dss, box_width = 2):
    no_datasets = len(dss)
    box_space = box_width / 4
    total_width = no_datasets * box_width + (no_datasets - 1) * box_space
    offsets = np.linspace(-total_width / 2 + box_width / 2, 
                        total_width / 2 - box_width / 2,
                        no_datasets)
    return offsets

def open_data(fh, mask, example, scale, offset, flatten = True):

    if not isinstance(example, type(None)):
        ds = PF.reproj_ds(fh, example)
    else:
        ds = gdal.Open(fh)

    if not isinstance(mask, type(None)):
        mask_ds = gdal.Open(mask)
        mask_array = mask_ds.GetRasterBand(1).ReadAsArray()
        array = ds.GetRasterBand(1).ReadAsArray()

        print("non-masked pixels (mask)", np.sum(np.isfinite(mask_array)))

        array[np.isnan(mask_array)] = np.nan
        array[mask_array != 1] = np.nan

        print("non-masked pixels (data)", np.sum(np.isfinite(array)))
    else:
        array = ds.GetRasterBand(1).ReadAsArray()

    ndv = ds.GetRasterBand(1).GetNoDataValue()

    # print("shape: {0}".format(array.shape))

    if flatten:
        flat_array = array.flatten().astype(np.float32)
    else:
        flat_array = array.astype(np.float32)

    flat_array[flat_array == ndv] = np.nan

    if not isinstance(scale, type(None)):
        flat_array *= scale

    if not isinstance(offset, type(None)):
        flat_array += offset

    return flat_array

def calc_pearson_correlation(arrays):
    # assert len(arrays) == 2

    x = arrays[0]

    rs = np.array([])

    for y in arrays[1:]:

        xmean = np.mean(x)
        ymean = np.mean(y)
        xdiff = x - xmean
        ydiff = y - ymean

        r = np.sum(xdiff * ydiff) / (np.sqrt(np.sum(xdiff**2)) * np.sqrt(np.sum(ydiff**2)))
        rs = np.append(rs, r)

    return rs

def calc_rmse(arrays):

    x = arrays[0]
    rmses = np.array([])

    for y in arrays[1:]:
        rmse = np.sqrt(np.mean((x - y)**2))
        rmses = np.append(rmses, rmse)
    
    return rmses

def calc_nash_sutcliffe(arrays):

    x = arrays[0]
    nses = np.array([])

    for y in arrays[1:]:

        # If x and y are identical return 1.0 (e.g. when all values in both x and y are zero)
        if np.all(x == y):
            nses = np.append(nses, 1.0)
        else:
            nse = 1 - (np.sum((x - y)**2) / np.sum((y - np.mean(y))**2))
            nses = np.append(nses, nse)

    return nses

def calc_rel_bias(arrays):

    x = arrays[0]
    rbs = np.array([])

    for y in arrays[1:]:

        rb = (np.sum(x) - np.sum(y)) / np.sum(x)

        rbs = np.append(rbs, rb)

    return rbs

def plot_honeycomb(dss, date, meta, output_fh = None, 
                    minmax = None, date_format = "%Y-%m", ax = None):

    if isinstance(ax, type(None)):
        plt.clf()
        fig = plt.figure(1)
        ax = fig.gca()

    arrays = list()

    for ds in dss:

        idx = ds["dates"].index(date)

        params = (ds["files"][idx], 
                 ds["masks"][idx],
                 ds["example"][idx],
                 ds["scale"][idx], 
                 ds["offset"][idx])

        array = open_data(*params)

        if not isinstance(minmax, type(None)):

            mini = np.nanpercentile(array, minmax[0])
            maxi = np.nanpercentile(array, minmax[1])

            min_mask = np.sum(array < mini)
            max_mask = np.sum(array > maxi)
            print(f"Masking {min_mask} pixels smaller than {mini}")
            array[array < mini] = np.nan
            print(f"Masking {max_mask} pixels greater than {maxi}")
            array[array > maxi] = np.nan

        arrays.append(array)

    orig_arrays = np.copy(arrays)

    mask = np.any([np.isnan(array) for array in arrays], axis = 0)
    arrays = np.array([array[~mask] for array in arrays])

    rs = calc_pearson_correlation(arrays)
    nses = calc_nash_sutcliffe(arrays)
    n = arrays[0].size

    hb = ax.hexbin(arrays[1], arrays[0], bins = "log")

    ax.set_facecolor(plt.cm.get_cmap('viridis')(0.0))

    ax.grid(color="k", linestyle=":")

    ax.set_title(meta["title"] + f"\n r: {rs[0]:.3f}, nse: {nses[0]:.3f}")
    # ax.set_title(meta["title"] )

    ax.set_xlabel(meta["labels"][1])
    ax.set_ylabel(meta["labels"][0])

    if not isinstance(output_fh, type(None)):
        plt.savefig(output_fh)

    return orig_arrays, arrays

def plot_boxplots(dss, meta, output_fh = None, box_width = 2, date_format = "%Y-%m", fig = None, ax = None, legend = True):

    offsets = determine_offsets(dss, box_width=box_width)

    if isinstance(fig, type(None)):
        plt.clf()
        plt.figure(1)
        fig = plt.gcf()

        width = (5/7) * len(dss[0]["files"])
        height = 4.2

        # fig.set_size_inches(3.6, 2.5)
        fig.set_size_inches(width, height)

    if isinstance(ax, type(None)):
        ax = fig.gca()

    for ds in dss:

        input_params = [ds["files"], ds["dates"], ds["dataset"],
                        ds["masks"], ds["scale"], ds["offset"]]
        
        for params in zip(*input_params):

            fh, date, ds_no = params[:3]

            if date in meta["exclude"] or fh in meta["exclude"]:
                continue

            flat_array = open_data(params)

            finite_array = flat_array[np.isfinite(flat_array)]

            i = date.toordinal()
            i += offsets[ds_no - 1]

            box = ax.boxplot(finite_array, positions = [i], widths = [box_width], 
                            showfliers = False, patch_artist = True, whis = (1, 99)
                            )

            clr = meta["colors"][ds_no - 1]
            box["medians"][0].set_color("black")
            box["medians"][0].set_linestyle("--")
            box["boxes"][0].set_linewidth(0.)
            box["boxes"][0].set_facecolor(clr)

            print("{0} processed.".format(date))

    xticks = [d.toordinal() for d in np.sort(np.unique(dss[0]["dates"])) if d not in meta["exclude"]]
    ax.set_xticks(xticks)
    xlabels = [d.date().strftime(date_format) for d in np.sort(np.unique(dss[0]["dates"])) if d not in meta["exclude"]]
    ax.set_xticklabels(xlabels, rotation = 20, ha = "right")

    ax.grid(zorder = 0)
    ax.set_facecolor("lightgray")
    if "xlabel" in meta.keys():
        ax.set_xlabel(meta["xlabel"])
    else:
        ax.set_xlabel("Dates")
    ax.set_ylabel(meta["ylabel"])
    ax.set_xlim([xticks[0] + offsets[0] * 2, xticks[-1] + offsets[-1] * 2])

    if legend:
        legend_elements = [Patch(facecolor = meta["colors"][ds_no - 1], label = meta["labels"][ds_no - 1]) for ds_no in np.arange(1, len(dss)+1)]
        ax.legend(handles = legend_elements, loc="best")
    ax.set_title(meta["title"])

    if not isinstance(output_fh, type(None)):
        fig.savefig(output_fh)

    return ax

def compare_two_tifs(tif1, tif2, example_tif, options1 = None, options2 = None, output = None, quant_unit = None):
    
    date = datetime.datetime(2021,1,1)

    ds1 = dict()
    ds1["files"] = [tif1]
    ds1["dates"] = [date]
    ds1["example"] = example_tif
    ds1["dataset"] = 1

    if not isinstance(options1, type(None)):
        for option, value in options1.items():
            ds1[option] = value

    ds2 = dict()
    ds2["files"] = [tif2]
    ds2["dates"] = [date]
    ds2["example"] = example_tif
    ds2["dataset"] = 2

    if not isinstance(options2, type(None)):
        for option, value in options2.items():
            ds2[option] = value

    meta = dict()
    meta["exclude"] = []
    meta["colors"] = ["r", "g"]
    meta["labels"] = [f"pyWaPOR {quant_unit[0]} [{quant_unit[1]}]", f"WaPOR {quant_unit[0]} [{quant_unit[1]}]"]
    meta["title"] = ""
    meta["ylabel"] = ""
    meta["xlabel"] = ""
    meta["sorter"] = "dates"

    dss = prepare_dss([ds1, ds2], meta["sorter"])

    ds = gdal.Open(example_tif)
    xsize = ds.RasterXSize
    ysize = ds.RasterYSize

    nrows = 2
    ncols = 3
    fig = plt.figure(1)
    fig.clf()
    fig.set_size_inches(11.69, 6.27) # 8.27

    axs = fig.subplots(nrows, ncols, sharex=False, sharey=False)
    axs = axs.flatten()

    plt.subplots_adjust(hspace=0.5, wspace = 0.3)

    orig_arrays, arrays = plot_honeycomb(dss, date, meta, ax = axs[2])

    

    mask = np.any([np.isnan(orig_arrays[0]), np.isnan(orig_arrays[1])], axis = 0)
    orig_arrays[0][mask] = np.nan
    orig_arrays[1][mask] = np.nan

    mini = np.nanpercentile(orig_arrays,  0.01)
    maxi = np.nanpercentile(orig_arrays, 99.99)

    axs[2].set_xlim([mini, maxi])
    axs[2].set_ylim([mini, maxi])

    axs[0].hist(arrays[0], bins = int((xsize*ysize)*0.00025))
    axs[0].set_title("pyWaPOR (n = {0})".format(arrays[0].size))
    axs[0].set_facecolor("lightgray")
    axs[0].grid(zorder = 0)
    axs[0].set_xlabel(f"{quant_unit[0]} [{quant_unit[1]}]")   
    axs[0].set_ylabel("Number of pixels [-]")
    axs[0].set_xlim([mini, maxi])

    axs[1].hist(arrays[1], bins = int((xsize*ysize)*0.00025))
    axs[1].set_title("WaPOR (n = {0})".format(arrays[1].size))
    axs[1].set_facecolor("lightgray")
    axs[1].grid(zorder = 0)
    axs[1].set_xlim([mini, maxi])
    axs[1].set_xlabel(f"{quant_unit[0]} [{quant_unit[1]}]")

    shw1 = axs[3].imshow(orig_arrays[0].reshape((ysize,xsize)), vmin = mini, vmax = maxi)
    shw2 = axs[4].imshow(orig_arrays[1].reshape((ysize,xsize)), vmin = mini, vmax = maxi)
    
    error_img = orig_arrays[0].reshape((ysize,xsize)) - orig_arrays[1].reshape((ysize,xsize))
    error_lim = np.nanpercentile(np.abs(error_img), 99)

    shw3 = axs[5].imshow(error_img, vmin = -1 * error_lim, vmax = error_lim, cmap = 'coolwarm')

    xax3 = axs[3].axes.get_xaxis()
    yax3 = axs[3].axes.get_yaxis()
    xax3 = xax3.set_visible(False)
    yax3 = yax3.set_visible(False)
    axs[3].set_title(f"pyWaPOR {quant_unit[0]}")

    xax4 = axs[4].axes.get_xaxis()
    yax4 = axs[4].axes.get_yaxis()
    xax4 = xax4.set_visible(False)
    yax4 = yax4.set_visible(False)
    axs[4].set_title(f"WaPOR {quant_unit[0]}")

    xax5 = axs[5].axes.get_xaxis()
    yax5 = axs[5].axes.get_yaxis()
    xax5 = xax5.set_visible(False)
    yax5 = yax5.set_visible(False)
    axs[5].set_title(f"pyWaPOR {quant_unit[0]} - WaPOR {quant_unit[0]}")

    cax3 = axs[3].inset_axes([0.00, -0.16, 1.0, 0.04], transform=axs[3].transAxes)
    cax4 = axs[4].inset_axes([0.00, -0.16, 1.0, 0.04], transform=axs[4].transAxes)
    cax5 = axs[5].inset_axes([0.00, -0.16, 1.0, 0.04], transform=axs[5].transAxes)

    fig.colorbar(shw1, cax = cax3, ax = axs[3], extend = "both", orientation = "horizontal")
    fig.colorbar(shw2, cax = cax4, ax = axs[4], extend = "both", orientation = "horizontal")
    fig.colorbar(shw3, cax = cax5, ax = axs[5], extend = "both", orientation = "horizontal")

    fig.suptitle(f'{os.path.split(tif1)[-1]} vs. {os.path.split(tif2)[-1]}')

    if not isinstance(output, type(None)):
        fig.savefig(output)

    return np.nanmean(orig_arrays[0]), np.nanmean(orig_arrays[1]), [orig_arrays[0].reshape((ysize,xsize)), orig_arrays[1].reshape((ysize,xsize))]

def plot_ts(ax, times, values, sources, styling):

    for source in np.unique(sources):

        if source == -1:
            continue

        xs = times[sources == source]
        ys = values[sources == source]

        if np.sum(np.isnan(ys)) == ys.size:
            continue

        c_styling = styling[source]

        ax.scatter(xs, ys, marker = c_styling[0], c = c_styling[1], label = c_styling[3], zorder = 10, alpha = float(c_styling[2]))
        ax.set_facecolor("lightgray")
        ax.grid(color="k", linestyle=":")
        ax.legend(loc = "best")
        ax.set_title("RAW Values")

    # ax.set_ylabel("NDVI [-]")

    return ax

def plot_composite_ts(ax, starts, ends, values, label = "Composite"):

    ax.bar(starts, values, (ends - starts), align = "edge", color = "lightblue", edgecolor = "darkblue", label = label)
    ax.set_facecolor("lightgray")
    ax.grid(color="k", linestyle=":")
    # ax.set_ylabel("NDVI [-]")
    ax.set_title("Composite Values")
    ax.legend(bbox_to_anchor=(1,1), loc="upper left")

def check_bb(ds, diagnostics_lon, diagnostics_lat):
    lon_lim = [float(ds.lon.min().values), 
                float(ds.lon.max().values)]
    lat_lim = [float(ds.lat.min().values), 
                float(ds.lat.max().values)]
    lon_test = np.all([diagnostics_lon >= lon_lim[0],
                       diagnostics_lon <= lon_lim[1]])
    lat_test = np.all([diagnostics_lat >= lat_lim[0],
                       diagnostics_lat <= lat_lim[1]])
    if not lon_test:
        print(f"WARNING: {diagnostics_lon} not in boundingbox ({lon_lim}).")
    if not lat_test:
        print(f"WARNING: {diagnostics_lat} not in boundingbox ({lat_lim}).")

def plot_composite(ds, diagnostics, out_folder = None):

    ds = ds.reindex({"y": ds.y.sortby("y"), 
                     "x": ds.x.sortby("x")})

    var = "_".join([x for x in ds.data_vars if "_values" in x][0].split("_")[0:-1])

    styling = dict()
    markers = ["*", "o", "v", "s", "*", "p", "h"]
    colors =  ["r", "g", "b", "y", "purple", "darkblue", "gray", "orange"]
    for i, source_name in ds[f"{var}_source"].attrs.items():
        i = int(i)
        styling[i] = (markers[i], colors[i], 1.0, source_name)
    styling[255] = (".", "k", 0.7, "Interp.")

    for point, (diagnostics_lat, diagnostics_lon) in diagnostics.items():

        ts = ds.sel(x = diagnostics_lon, 
                    y = diagnostics_lat, method="nearest")

        times = ts.time.values
        ndvis = ts[f"{var}_values"].values
        sources = ts[f"{var}_source"].values
        ndvi_composites = ts[var].values

        epoch_starts = ts.time_bins.values
        epoch_ends = np.append(epoch_starts[1:], np.datetime64(ds.bin_end))

        fig = plt.figure(1)
        fig.clf()
        fig.set_size_inches(9, 5)
        ax = fig.gca()

        plot_ts(ax, times, ndvis, sources, styling)
        plot_composite_ts(ax, epoch_starts, epoch_ends, ndvi_composites, f"Composite ({ds.attrs['comp_type']})")

        ax.set_ylabel(f"{var}")
        ax.set_title(f"{var} at {point}")

        fig.autofmt_xdate(bottom=0.2, rotation=30, ha='right')
        fig.suptitle(f"{diagnostics_lon} °E, {diagnostics_lat} °N")
        fig.subplots_adjust(right=0.75)

        if not isinstance(out_folder, type(None)):
            if not os.path.exists(out_folder):
                os.makedirs(out_folder)
            fn = f"{point}_{var}.png"
            fig.savefig(os.path.join(out_folder, fn))

def plot_tif(tif_file, quantity = None, unit = None):
    ds = gdal.Open(tif_file)
    array = ds.GetRasterBand(1).ReadAsArray()
    ndv = ds.GetRasterBand(1).GetNoDataValue()

    array[array == ndv] = np.nan

    mini = np.nanpercentile(array, 5)
    maxi = np.nanpercentile(array, 95)

    fn = os.path.split(tif_file)[-1]

    plt.imshow(array, vmin = mini, vmax = maxi)
    plt.colorbar(label = f"{quantity} {unit}", extend = "both")
    plt.title(fn)
    plt.gca().set_facecolor("lightgray")

def prettyprint(d, indent=0):
   for key, value in d.items():
      print('\t' * indent + str(key))
      if isinstance(value, dict):
         prettyprint(value, indent+1)
      else:
         print('\t' * (indent) + " " + str(value))

def plot_img(ax, array, title = "", var_name = "", cmap = "viridis", cb_limits = None):
    if cmap == "coolwarm":
        max_error = np.nanpercentile(np.abs(array), 99)
        vmin = -max_error
        vmax = max_error
    else:
        vmin = None
        vmax = None
    if not isinstance(cb_limits, type(None)):
        vmin = cb_limits[0]
        vmax = cb_limits[1]
    img0 = ax.imshow(array, cmap = cmap, vmin = vmin, vmax = vmax)
    cax0 = ax.inset_axes([0.00, -0.20, 1.0, 0.04], transform=ax.transAxes)
    ax.get_figure().colorbar(img0, ax = ax, cax = cax0, extend = "both", 
                orientation = "horizontal", label = var_name)
    ax.set_title(title)

def plot_hexbin(ax, arrays, xlabel = "", ylabel = "", title = ""):
    minmax = [np.min(arrays), np.max(arrays)]
    ax.plot(minmax, minmax, ":k", label = "1:1")
    ax.legend()
    hb = ax.hexbin(*arrays, bins = "log")
    ax.set_facecolor(plt.cm.get_cmap('viridis')(0.0))
    ax.get_figure().colorbar(hb, label = "Number of pixels [-]")
    ax.set_xlim(minmax)
    ax.set_ylim(minmax)
    ax.set_xlabel(xlabel)
    ax.set_ylabel(ylabel)
    ax.set_title(title)

# import geoviews as gv
# import holoviews as hv
# hv.extension('bokeh')

# def make_geoview(ds):
#     da = ds.to_array("variable")

#     gv_ds = gv.Dataset(da, vdims = ["data"])

#     out = gv_ds.to(gv.Image, ["x", "y"]).opts(colorbar = True, width = 500, height = 500)

#     return out


# styling[999] = ("P", "orange", 1.0, "-")
# styling[0] = (".", "k", 0.7, "Interp.")
# sources = {v[3]: k for k, v in styling.items()}



import os
import rasterio
import rioxarray.merge
import tempfile
import warnings
import requests
import copy
import functools
import urllib.parse
import xarray as xr
import multiprocessing
from joblib import Parallel, delayed
import numpy as np
from pywapor.general.logger import log
from rasterio.crs import CRS
from bs4 import BeautifulSoup
from urllib.parse import urlsplit, urlunsplit
from pywapor.general.processing_functions import open_ds, is_corrupt_or_empty
from pywapor.enhancers.apply_enhancers import apply_enhancers
from pywapor.collect.protocol.crawler import download_url, download_urls
from pywapor.general.processing_functions import save_ds, process_ds, remove_ds

def make_opendap_url(base_url, order):
    """_summary_

    Parameters
    ----------
    base_url : str
        URL to OPeNDAP database
    order : dict
        Keys are OPeNDAP variables, values are dictionaries specifying
        the indices to download for a dimension. Give an empty dictionary
        to not do any slicing.

    Returns
    -------
    str
        URL to download a subset.

    Example
    -------
    base_url = "https://ladsweb.modaps.eosdis.nasa.gov/opendap/RemoteResources/laads/allData/5200/VNP02IMG/2023/060/VNP02IMG.A2023060.1106.002.2023061191604.nc"
    order = {
            '/observation_data/I05_quality_flags': {
                'number_of_lines': [1610, 1711],
                'number_of_pixels': [2920, 3013]},
            '/observation_data/I05': {
                'number_of_lines': [1610, 1711],
                'number_of_pixels': [2920, 3013]},
            '/observation_data/I05_brightness_temperature_lut': {}
            }
    """

    dims_name_order = {
        "/geolocation_data/longitude": ["number_of_lines", "number_of_pixels"],
        "/geolocation_data/latitude": ["number_of_lines", "number_of_pixels"],
        "/observation_data/I05_quality_flags": ["number_of_lines", "number_of_pixels"],
        "/observation_data/I05": ["number_of_lines", "number_of_pixels"],
        "/observation_data/I05_brightness_temperature_lut": ["number_of_LUT_values"],
        "/geophysical_data/Integer_Cloud_Mask": ["number_of_lines", "number_of_pixels"],
    }

    all_var_order_strs = []
    suffix = ".dap.nc4"
    for var, subset in order.items():
        # TODO auto lookup dimension order for var if its not predefined.
        all_dims = dims_name_order[var]
        all_subset_strs = []
        for dim in all_dims:
            subset_str = urllib.parse.quote(f"[{subset[dim][0]}:1:{subset[dim][1]}]") if dim in subset.keys() else urllib.parse.quote("[]")
            all_subset_strs.append(subset_str)
        var_order_str = f"{var}{''.join(all_subset_strs)}"
        all_var_order_strs.append(var_order_str)
    order_str = ';'.join(all_var_order_strs)
    url = base_url + suffix + "?dap4.ce=" + order_str
    return url

def download(fp, product_name, coords, variables, post_processors, 
                fn_func, url_func, un_pw = None, tiles = None,  
                data_source_crs = None, parallel = False, spatial_tiles = True, 
                request_dims = True, timedelta = None):
    """Download data from a OPENDaP server.

    Parameters
    ----------
    fp : str
        Path to file in which to download.
    product_name : str
        Name of product.
    coords : dict
        Coordinate names and boundaries.
    variables : dict
        Keys are variable names, values are additional settings.
    post_processors : dict
        Processors to apply to specific variables.
    url_func : function
        Function that takes `product_name` as input and return a url.
    un_pw : tuple, optional
        Username and password to use, by default None.
    tiles : list, optional
        Tiles to download, by default None.
    data_source_crs : rasterio.CRS.crs, optional
        CRS of datasource, by default None.
    parallel : bool, optional
        Download files in parallel (currently not implemented), by default False.
    spatial_tiles : bool, optional
        Whether the tiles are spatial or temporal, by default True.
    request_dims : bool, optional
        Include dimension settings in the OPENDaP request, by default True.
    timedelta : datetime.datetime.timedelta, optional
        Shift the time coordinates by tdelta, by default None.

    Returns
    -------
    xr.Dataset
        Dataset with downloaded data.
    """

    folder = os.path.split(fp)[0]

    # Create selection object.
    selection = create_selection(coords, target_crs = data_source_crs)

    # Make output filepaths, should be same length as `urls`.
    fps = [os.path.join(folder, fn_func(product_name, x)) for x in tiles]

    # Make data request URLs.
    session = start_session(url_func(product_name, tiles[0]), selection, un_pw)
    if spatial_tiles:
        idxss = [find_idxs(url_func(product_name, x), selection, session) for x in tiles]
        urls = [create_url(url_func(product_name, x), idxs, variables, request_dims = request_dims) for x, idxs in zip(tiles, idxss)]
    else:
        idxs = find_idxs(url_func(product_name, tiles[0]), selection, session)
        urls = [create_url(url_func(product_name, x), idxs, variables, request_dims = request_dims) for x in tiles]

    # Download data.
    files = download_urls(urls, "", session, fps = fps, parallel = parallel)

    # Merge spatial tiles.
    coords_ = {k: [v[0], selection[v[0]]] for k,v in coords.items()}
    if spatial_tiles:
        dss = [process_ds(xr.open_dataset(x, decode_coords = "all"), coords_, variables, crs = data_source_crs) for x in files]
        ds = rioxarray.merge.merge_datasets(dss)
    else:
        dss = files
        ds = xr.concat([xr.open_dataset(x, decode_coords="all") for x in files], dim = "time")
        ds = process_ds(ds, coords_, variables, crs = data_source_crs)

    # Reproject if necessary.
    if ds.rio.crs.to_epsg() != 4326:
        ds = ds.rio.reproject(CRS.from_epsg(4326))

    ds = ds.rio.clip_box(coords["x"][1][0], coords["y"][1][0], coords["x"][1][1], coords["y"][1][1])

    # Apply product specific functions.
    ds = apply_enhancers(post_processors, ds)

    if isinstance(timedelta, np.timedelta64):
        ds["time"] = ds["time"] + timedelta

    # Remove unrequested variables.
    ds = ds[list(post_processors.keys())]
    
    # Save final output.
    ds.attrs = {}
    for var in ds.data_vars: # NOTE Keeping band attributes can cause problems when 
        # opening the data using rasterio (in reproject_chunk), see https://github.com/rasterio/rasterio/discussions/2751
        ds[var].attrs = {}

    ds = save_ds(ds, fp, encoding = "initiate", label = "Saving merged data.")

    # Remove temporary files.
    if not isinstance(dss, type(None)):
        for x in dss:
            remove_ds(x)

    return ds

def find_idxs(base_url, selection, session):
    fp = tempfile.NamedTemporaryFile(suffix=".nc").name
    url_coords = base_url + urllib.parse.quote(",".join(selection.keys()))
    fp = download_url(url_coords, fp, session, waitbar = False)
    ds = xr.open_dataset(fp, decode_coords = "all")
    idxs = dict()
    for k, v in selection.items():
        all_idxs = np.where((ds[k] >= v[0]) & (ds[k] <= v[1]))[0]
        if len(all_idxs) == 0:
            msg = f"--> No data found for `{k}` in range [{v[0]}, ...,{v[1]}]. Data is available between [{ds[k][0].values}, ..., {ds[k][-1].values}]."
            log.error(msg)
            raise ValueError(msg)
        idxs[k] = [np.min(all_idxs), np.max(all_idxs)]
    return idxs

def create_url(base_url, idxs, variables, request_dims = True):
    if request_dims:
        dims = [f"{k}[{v[0]}:{v[1]}]" for k, v in idxs.items()]
    else:
        dims = []
    varis = [f"{k}{''.join([f'[{idxs[dim][0]}:{idxs[dim][1]}]' for dim in v[0]])}" for k, v in variables.items()]
    url = base_url + urllib.parse.quote(",".join(dims + varis))
    return url

def start_session(base_url, selection, un_pw = [None, None]):
    if un_pw == [None, None]:
        warnings.filterwarnings("ignore", "password was not set. ")
    url_coords = base_url + urllib.parse.quote(",".join(selection.keys()))
    vrfy = {"NO": False, "YES": True}.get(os.environ.get("PYWAPOR_VERIFY_SSL", "YES"), True)
    session = setup_session('https://urs.earthdata.nasa.gov', username = un_pw[0], password = un_pw[1], check_url=url_coords, verify=vrfy)
    return session

def setup_session(uri,
                  username=None,
                  password=None,
                  check_url=None,
                  session=None,
                  verify=True,
                  username_field='username',
                  password_field='password'):
    '''
    A general function to set-up requests session with cookies
    using beautifulsoup and by calling the right url.
    '''

    if session is None:
        # Connections must be closed since some CAS
        # will cough when connections are kept alive:
        headers = [
            # ('User-agent', 'pydap/{}'.format(lib.__version__)),
                   ('Connection', 'close')]
        session = requests.Session()
        session.headers.update(headers)

    if uri is None:
        return session

    if not verify:
        verify_flag = session.verify
        session.verify = False

    if isinstance(uri, str):
        url = uri
    else:
        url = uri(check_url)

    if password is None or password == '':
        warnings.warn('password was not set. '
                      'this was likely unintentional '
                      'but will result is much fewer datasets.')
        if not verify:
            session.verify = verify_flag
        return session

    # Allow for several subsequent security layers:
    full_url = copy.copy(url)
    if isinstance(full_url, list):
        url = full_url[0]

    with warnings.catch_warnings():

        response = soup_login(session, url, username, password,
                              username_field=username_field,
                              password_field=password_field)

        # If there are further security levels.
        # At the moment only used for CEDA OPENID:
        if (isinstance(full_url, list) and
           len(full_url) > 1):
            for url in full_url[1:]:
                response = soup_login(session, response.url,
                                      username, password,
                                      username_field=None,
                                      password_field=None)
        response.close()

        if check_url:
            if (username is not None and
               password is not None):
                res = session.get(check_url, auth=(username, password))
                if res.status_code == 401:
                    res = session.get(res.url, auth=(username, password))
                res.close()
            raise_if_form_exists(check_url, session)

    if not verify:
        session.verify = verify_flag
    return session

def raise_if_form_exists(url, session):
    """
    This function raises a UserWarning if the link has forms
    """

    user_warning = ('Navigate to {0}, '.format(url) +
                    'login and follow instructions. '
                    'It is likely that you have to perform some one-time '
                    'registration steps before acessing this data.')

    resp = session.get(url)
    soup = BeautifulSoup(resp.content, 'lxml')
    if len(soup.select('form')) > 0:
        raise UserWarning(user_warning)


def soup_login(session, url, username, password,
               username_field='username',
               password_field='password'):
    resp = session.get(url)

    soup = BeautifulSoup(resp.content, 'lxml')
    login_form = soup.select('form')[0]

    def get_to_url(current_url, to_url):
        split_current = urlsplit(current_url)
        split_to = urlsplit(to_url)
        comb = [val2 if val1 == '' else val1
                for val1, val2 in zip(split_to, split_current)]
        return urlunsplit(comb)
    to_url = get_to_url(resp.url, login_form.get('action'))

    session.headers['Referer'] = resp.url

    payload = {}
    if username_field is not None:
        if len(login_form.findAll('input', {'name': username_field})) > 0:
            payload.update({username_field: username})

    if password_field is not None:
        if len(login_form.findAll('input', {'name': password_field})) > 0:
            payload.update({password_field: password})
        else:
            # If there is no password_field, it might be because
            # something should be handled in the browser
            # for the first attempt. This is common when using
            # pydap with the ESGF for the first time.
            raise Exception('Navigate to {0}. '
                            'If you are unable to '
                            'login, you must either '
                            'wait or use authentication '
                            'from another service.'
                            .format(url))

    # Replicate all other fields:
    for input in login_form.findAll('input'):
        if (input.get('name') not in payload and
           input.get('name') is not None):
            payload.update({input.get('name'): input.get('value')})

    # Remove other submit fields:
    submit_type = 'submit'
    submit_names = [input.get('name') for input
                    in login_form.findAll('input', {'type': submit_type})]
    for input in login_form.findAll('input', {'type': submit_type}):
        if ('submit' in submit_names and
           input.get('name').lower() != 'submit'):
            payload.pop(input.get('name'), None)

    return session.post(to_url, data=payload)

def download_xarray(url, fp, coords, variables, post_processors, 
                    data_source_crs = None, timedelta = None):
    """Download a OPENDaP dataset using xarray directly.

    Parameters
    ----------
    url : str
        URL to dataset.
    fp : str
        Path to file to download into.
    coords : dict
        Coordinates to request.
    variables : dict
        Variables to request.
    post_processors : dict
        Additional functions to apply to variables.
    data_source_crs : rasterio.CRS.crs, optional
        CRS of the data source, by default None.
    timedelta : datetime.datetime.timedelta, optional
        Shift the time coordinates by tdelta, by default None.

    Returns
    -------
    xr.Dataset
        Downloaded dataset.
    """

    warnings.filterwarnings("ignore", category=xr.SerializationWarning)
    online_ds = xr.open_dataset(url, decode_coords="all")

    # Define selection.
    selection = create_selection(coords, target_crs = data_source_crs)

    # Make the selection on the remote.
    online_ds = online_ds.sel({k: slice(*v) for k, v in selection.items()})

    # Rename variables and assign crs.
    online_ds = process_ds(online_ds, coords, variables, crs = data_source_crs)

    def download_chunk(da, fp):
        time_str = str(da.isel({"time":0}).time.dt.strftime("%Y%m%d_%H%M%S").values)
        fp_ = fp.replace(".nc", f"_{time_str}_{da.time.size}.nc")
        corrupt = False
        if os.path.isfile(fp_):
            corrupt = is_corrupt_or_empty(fp_)
            if corrupt:
                log.info(f"--> Removing corrupt or empty file `{os.path.split(fp_)[-1]}`.")
                remove_ds(fp_)
            else:
                log.info(f"--> Opening existing file `{os.path.split(fp_)[-1]}`.")
                ds = open_ds(fp_)
        if corrupt or not os.path.isfile(fp_):
            log.info(f"--> Downloading file `{os.path.split(fp_)[-1]}`.")
            da.to_netcdf(fp_)
            ds = open_ds(fp_)
        return ds
    
    n_jobs = max(1, multiprocessing.cpu_count() - 1)
    block_size = max(10, min(500, int(np.ceil(online_ds.time.size / n_jobs))))

    chunks = [online_ds.isel({"time": slice(i*block_size, block_size*(i+1))}) for i in range(int(np.ceil(online_ds.time.size / block_size)))]
    n_jobs = min(n_jobs, len(chunks))

    parallel = True
    if parallel:
        out = Parallel(n_jobs=n_jobs)(delayed(download_chunk)(subds, fp) for subds in chunks)
    else:
        out = list(map(download_chunk, chunks))
    
    ds = xr.concat(out, dim="time")

    # Apply product specific functions.
    ds = apply_enhancers(post_processors, ds)

    if isinstance(timedelta, np.timedelta64):
        ds["time"] = ds["time"] + timedelta

    # Save final output
    out = save_ds(ds, fp, encoding = "initiate", label = "Saving netCDF.")

    for x in out:
        remove_ds(x)

    return out

def create_selection(coords, target_crs = None, source_crs = CRS.from_epsg(4326)):
    """Create a dictionary that can be given to `xr.Dataset.sel`.

    Parameters
    ----------
    coords : dict
        Dictionary describing the different dimensions over which to select. Possible keys
        are "x" for latitude, "y" for longitude and "t" for time, but other selections
        keys are also allowed (e.g. so select a band). Values are tuples with the first
        value the respective dimension names in the `ds` and the second value the selector.
    target_crs : rasterio.crs.CRS, optional
        crs of the dataset on which the selection will be applied, by default None.
    source_crs : rasterio.crs.CRS, optional
        crs of the `x` and `y` limits in `coords`, by default `epsg:4326`.

    Returns
    -------
    dict
        Dimension names with slices to apply to each dimension.
    """
    selection = {}

    if not isinstance(target_crs, type(None)):
        bounds = rasterio.warp.transform_bounds(source_crs, target_crs, 
                                                coords["x"][1][0], 
                                                coords["y"][1][0], 
                                                coords["x"][1][1], 
                                                coords["y"][1][1])
    else:
        bounds = [coords["x"][1][0], coords["y"][1][0], coords["x"][1][1], coords["y"][1][1]]
    
    selection[coords["x"][0]] = [bounds[0], bounds[2]]
    selection[coords["y"][0]] = [bounds[1], bounds[3]]

    if "t" in coords.keys():
        selection[coords["t"][0]] = [np.datetime64(t) for t in coords["t"][1]]

    for name, lim in coords.values():
        if name not in selection.keys():
            selection[name] = lim

    return selection


if __name__ == "__main__":
    ...

    # from pywapor.collect.protocol.projections import get_crss

    # url = 'https://opendap.nccs.nasa.gov/dods/GEOS-5/fp/0.25_deg/assim/tavg1_2d_slv_Nx'

    # coords = {'x': ['lon', [91.62463492436828, 92.27825666088243]],
    #             'y': ['lat', [21.719219468262693, 22.24391208383405]],
    #             't': ['time', ['2022-01-01', '2022-12-31']]}
    
    # data_source_crs = get_crss("WGS84")

    # variables = {'u10m': [('time', 'lat', 'lon'), 'u10m'],
    #             'v10m': [('time', 'lat', 'lon'), 'v10m'],
    #             'qv2m': [('time', 'lat', 'lon'), 'qv'],
    #             'slp': [('time', 'lat', 'lon'), 'p_air_0'],
    #             'ps': [('time', 'lat', 'lon'), 'p_air'],
    #             't2m': [('time', 'lat', 'lon'), 't_air'],
    #             'tqv': [('time', 'lat', 'lon'), 'wv']}


    # fp = "/Users/hmcoerver/Local/geos_test/GEOS5/tavg1_2d_slv_Nx.nc"

    # from pywapor.general.logger import adjust_logger
    # adjust_logger(True, "/Users/hmcoerver/Local/geos_test", "INFO")